# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import collections
import logging
import math
import re
import sys

from builtins import object
from concurrent.futures import Future
from functools import partial

import tornado
from future.utils import PY2

from tornado.concurrent import Future as tornado_Future
from tornado.gen import Return

from katcp import Message, inspecting_client, resource
from katcp.core import (AsyncCallbackEvent, AsyncEvent, AsyncState, AttrDict,
                        DefaultAttrDict, LatencyTimer, log_future_exceptions,
                        steal_docstring_from, until_any, until_some)
# TODO NM 2017-04-13 Importing IOLoopThreadwrapper here for backwards
# compatibility, user code should be changed to import it from the more logical
# katcp.ioloop_manager module.
from katcp.ioloop_manager import IOLoopThreadWrapper, ThreadSafeMethodAttrWrapper
from katcp.resource import KATCPReply, KATCPSensorError

log = logging.getLogger(__name__)


def _normalise_request_name_set(reqs):
    return set(resource.escape_name(r) for r in reqs)

def transform_future(transformation, future):
    """Returns a new future that will resolve with a transformed value.

    Takes the resolution value of `future` and applies "transformation(\*future.result())"
    to it before setting the result of the new future with the transformed value. If
    future() resolves with an exception, it is passed through to the new future.

    Assumes `future` is a tornado Future.

    """
    new_future = tornado_Future()
    def _transform(f):
        assert f is future
        if f.exc_info() is not None:
            new_future.set_exc_info(f.exc_info())
        else:
            try:
                new_future.set_result(transformation(f.result()))
            except Exception:
                # An exception here indicates that the transformation was unsuccessful
                new_future.set_exc_info(sys.exc_info())

    future.add_done_callback(_transform)
    return new_future

@tornado.gen.coroutine
def list_sensors(parent_class, sensor_items, filter, strategy, status,
                 use_python_identifiers, tuple, refresh):
    """Helper for implementing :meth:`katcp.resource.KATCPResource.list_sensors`

    Parameters
    ----------

    sensor_items : tuple of sensor-item tuples
        As would be returned the items() method of a dict containing KATCPSensor objects
        keyed by Python-identifiers.
    parent_class: KATCPClientResource or KATCPClientResourceContainer
        Is used for prefix calculation
    Rest of parameters as for :meth:`katcp.resource.KATCPResource.list_sensors`
    """
    filter_re = re.compile(filter)
    found_sensors = []
    none_strat = resource.normalize_strategy_parameters('none')
    sensor_dict = dict(sensor_items)
    for sensor_identifier in sorted(sensor_dict.keys()):
        sensor_obj = sensor_dict[sensor_identifier]
        search_name = (sensor_identifier if use_python_identifiers
                       else sensor_obj.name)
        name_match = filter_re.search(search_name)
        # Only include sensors with strategies
        strat_match = not strategy or sensor_obj.sampling_strategy != none_strat
        if name_match and strat_match:
            if refresh:
                # First refresh the sensor reading
                yield sensor_obj.get_value()
            # Determine the sensorname prefix:
            # parent_name. except for aggs when in KATCPClientResourceContinaer
            prefix = ""
            if isinstance(parent_class, KATCPClientResourceContainer):
                if sensor_obj.name.startswith("agg_"):
                    prefix = ""
                else:
                    prefix = sensor_obj.parent_name + "."
            if not status or (sensor_obj.reading.status in status):
                # Only include sensors of the given status
                if tuple:
                    # (sensor.name, sensor.value, sensor.value_seconds, sensor.type, sensor.units, sensor.update_seconds, sensor.status, strategy_and_params)
                    found_sensors.append((
                        prefix+sensor_obj.name,
                        sensor_obj.reading.value,
                        sensor_obj.reading.timestamp,
                        sensor_obj.type,
                        sensor_obj.units,
                        sensor_obj.reading.received_timestamp,
                        sensor_obj.reading.status,
                        sensor_obj.sampling_strategy
                        ))
                else:
                    found_sensors.append(resource.SensorResultTuple(
                        object=sensor_obj,
                        name=prefix+sensor_obj.name,
                        python_identifier=sensor_identifier,
                        description=sensor_obj.description,
                        units=sensor_obj.units,
                        type=sensor_obj.type,
                        reading=sensor_obj.reading))
    raise tornado.gen.Return(found_sensors)


class ReplyWrappedInspectingClientAsync(inspecting_client.InspectingClientAsync):
    """Adds wrapped_request() method that wraps reply in a KATCPReply """

    reply_wrapper = staticmethod(lambda x : KATCPReply(*x))

    def wrapped_request(self, request, *args, **kwargs):
        """Create and send a request to the server.

        This method implements a very small subset of the options
        possible to send an request. It is provided as a shortcut to
        sending a simple wrapped request.

        Parameters
        ----------
        request : str
            The request to call.
        *args : list of objects
            Arguments to pass on to the request.

        Keyword Arguments
        -----------------
        timeout : float or None, optional
            Timeout after this amount of seconds (keyword argument).
        mid : None or int, optional
            Message identifier to use for the request message. If None, use either
            auto-incrementing value or no mid depending on the KATCP protocol version
            (mid's were only introduced with KATCP v5) and the value of the `use_mid`
            argument. Defaults to None.
        use_mid : bool
            Use a mid for the request if True.

        Returns
        -------
        future object that resolves with the
        :meth:`katcp.client.DeviceClient.future_request` response wrapped in
        self.reply_wrapper

        Example
        -------

        ::

        wrapped_reply = yield ic.simple_request('help', 'sensor-list')

        """
        f = tornado_Future()
        try:
            use_mid = kwargs.get('use_mid')
            timeout = kwargs.get('timeout')
            mid = kwargs.get('mid')
            msg = Message.request(request, *args, mid=mid)
        except Exception:
            f.set_exc_info(sys.exc_info())
            return f
        return transform_future(self.reply_wrapper,
                                self.katcp_client.future_request(msg, timeout, use_mid))

class KATCPClientResource(resource.KATCPResource):
    """Class managing a client connection to a single KATCP resource

    Inspects the KATCP interface of the resources, exposing sensors and requests as per
    the :class:`katcp.resource.KATCPResource` API. Can also operate without exposing
    """

    MAX_LOOP_LATENCY = 0.03
    """
    When doing potentially tight loops in coroutines yield tornado.gen.moment
    after this much time. This is a suggestion for methods to use.
    """

    @property
    def state(self):
        return self._state.state

    @property
    def controlled(self):
        return self._controlled

    @property
    def req(self):
        return self._req

    @property
    def sensor(self):
        return self._sensor

    @property
    def address(self):
        return self._address

    @property
    def host(self):
        return self._address[0]

    @property
    def port(self):
        return self._address[1]

    @property
    def address_string(self):
        return "{}:{}".format(*self._address)

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def parent(self):
        return self._parent

    @property
    def parent_name(self):
        return self._parent.name

    @property
    def children(self):
        return {}

    @property
    def synced(self):
        return self.state == 'synced'

    @property
    def versions(self):
        try:
            ic = self._inspecting_client
        except AttributeError:
            raise RuntimeError('Cannot find inspecting client, have you called start()?')
        return ic.katcp_client.versions

    @property
    def last_connect_time(self):
        try:
            ic = self._inspecting_client
        except AttributeError:
            raise RuntimeError('Cannot find inspecting client, have you called start()?')
        return ic.katcp_client.last_connect_time

    @property
    def dummy_unknown_requests(self):
        return self._dummy_unknown_requests

    def __init__(self, resource_spec, parent=None, logger=log):
        """Initialise resource with given specification

        Parameters
        ----------
        resource_spec : dict with resource specifications. Keys:
          name : str
              Name of the resource.
          description : str, optional
              Description of the resource.
          address : (host, port), host as str, port as int
          always_allowed_requests : seq of str,
              KACTP requests that are always allowed, even when the resource is not
              controlled. '-' and '_' will be treated equivalently.
          always_excluded_requests : seq of str,
              KACTP requests that are never allowed, even if the resource is
              controlled. Overrides requests in `always_allowed_requests`. '-' and '_'
              will be treated equivalently.
          controlled : bool, default: False
              True if control over the device (i.e. KATCP requests) is to be exposed.
          auto_reconnect : bool, default: True
              If True, auto-reconnect should the network connection be closed.
          auto_reconnect_delay : float seconds. Default : 0.5s
              Delay between reconnection retries.
          dummy_unknown_requests : bool. Default : False
              If true, provide dummy request functions for any unknown requests. Can be
              used as a rough simulation of a device for testing when some requests are
              not available.
          preset_protocol_flags : :class:`katcp.core.ProtocolFlags` instance
              Assume these protocol settings and ignore the server's
              #katcp-protocol informs.
          # TODO(NM) 'keep', ie. katcorelib behaviour where requests / sensors never
          # disappear even if the device looses them. Or was it only sensors? Should look
          # at katcorelib

          # TODO, not implemented, proposed below for light non-inspecting mode

          inspect : bool, default : True
              Inspect the resource's KATCP interface for sensors and requests
          assumed_requests : ...
          assumed_sensors : ...

        parent : :class:`KATCPResource` or None
            Parent KATCPResource object if this client is a child in a resource hierarchy.

        logger : object, optional
           Python Logger object to log to. Default is the module logger

        """
        super(KATCPClientResource, self).__init__()
        self._address = resource_spec['address']
        self._name = resource_spec['name']
        self._description = resource_spec.get('description', '')
        self.always_allowed_requests = _normalise_request_name_set(
            resource_spec.get('always_allowed_requests', set()) )
        self.always_excluded_requests = _normalise_request_name_set(
            resource_spec.get('always_excluded_requests', set()) )
        self._controlled = resource_spec.get('controlled', False)
        self.auto_reconnect = resource_spec.get('auto_reconnect', True)
        self.auto_reconnect_delay = resource_spec.get('auto_reconnect_delay', 0.5)
        self._sensor_strategy_cache = {}
        self._sensor_listener_cache = collections.defaultdict(list)
        self._logger = logger
        self._parent = parent
        self._ioloop_set_to = None
        self._sensor = AttrDict()
        self._dummy_unknown_requests = bool(resource_spec.get('dummy_unknown_requests'))
        if self._dummy_unknown_requests:
            DummyRequest = partial(
                resource.KATCPDummyRequest,
                {'name':'dummy', 'description': 'No help for dummies',
                 'timeout_hint': None})
            self._req = DefaultAttrDict(DummyRequest)
        else:
            self._req = AttrDict()

        # Save the pop() / items() methods in case a sensor/request with the same name is
        # added
        self._preset_protocol_flags = resource_spec.get('preset_protocol_flags')
        self._state = AsyncState(("disconnected", "syncing", "synced"))
        self._connected = AsyncCallbackEvent(self._update_state)
        self._sensors_synced = AsyncCallbackEvent(self._update_state)
        self._requests_synced = AsyncCallbackEvent(self._update_state)

    def is_connected(self):
        """Indication of the connection state

        Returns True if state is not "disconnected", i.e "syncing" or "synced"
        """
        return not self.state == 'disconnected'

    def until_state(self, state, timeout=None):
        """Future that resolves when a certain client state is attained

        Parameters
        ----------

        state : str
            Desired state, one of ("disconnected", "syncing", "synced")
        timeout: float
            Timeout for operation in seconds.
        """
        return self._state.until_state(state, timeout=timeout)

    def wait_connected(self, timeout=None):
        """Future that resolves when the state is not 'disconnected'."""
        return self._state.until_state_in('syncing', 'synced', timeout=timeout)

    def _update_state(self, _flag=None):
        # Update self._state, optional _flag parameter is ignored to be compatible with
        # AsyncCallbackEvent
        if not self._connected.isSet():
            self._state.set_state('disconnected')
        else:
            if self._sensors_synced.isSet() and self._requests_synced.isSet():
                self._state.set_state('synced')
            else:
                self._state.set_state('syncing')

    def set_ioloop(self, ioloop=None):
        """Set the tornado ioloop to use

        Defaults to tornado.ioloop.IOLoop.current() if set_ioloop() is not called or if
        ioloop=None. Must be called before start()
        """
        self._ioloop_set_to = ioloop

    def start(self):
        """Start the client and connect"""
        # TODO (NM 2015-03-12) Some checking to prevent multiple calls to start()
        host, port = self.address
        ic = self._inspecting_client = self.inspecting_client_factory(
            host, port, self._ioloop_set_to)
        self.ioloop = ic.ioloop
        if self._preset_protocol_flags:
            ic.preset_protocol_flags(self._preset_protocol_flags)
        ic.katcp_client.auto_reconnect_delay = self.auto_reconnect_delay
        ic.set_state_callback(self._inspecting_client_state_callback)
        ic.request_factory = self._request_factory
        self._sensor_manager = KATCPClientResourceSensorsManager(
            ic, self.name, logger=self._logger)
        ic.handle_sensor_value()
        ic.sensor_factory = self._sensor_manager.sensor_factory

        # Steal some methods from _sensor_manager
        self.reapply_sampling_strategies = self._sensor_manager.reapply_sampling_strategies
        log_future_exceptions(self._logger, ic.connect())

    def inspecting_client_factory(self, host, port, ioloop_set_to):
        """Return an instance of :class:`ReplyWrappedInspectingClientAsync` or similar

        Provided to ease testing. Dynamically overriding this method after instantiation
        but before start() is called allows for deep brain surgery. See
        :class:`katcp.fake_clients.fake_inspecting_client_factory`

        """
        return ReplyWrappedInspectingClientAsync(
            host, port, ioloop=ioloop_set_to, auto_reconnect=self.auto_reconnect)

    def until_synced(self, timeout=None):
        """Convenience method to wait (with Future) until client is synced"""
        return self._state.until_state('synced', timeout=timeout)

    def until_not_synced(self, timeout=None):
        """Convenience method to wait (with Future) until client is not synced"""
        not_synced_states = [state for state in self._state.valid_states
                             if state != 'synced']
        not_synced_futures = [self._state.until_state(state)
                              for state in not_synced_states]
        return until_any(*not_synced_futures, timeout=timeout)

    @steal_docstring_from(resource.KATCPResource.list_sensors)
    def list_sensors(self, filter="", strategy=False, status="",
                     use_python_identifiers=True, tuple=False, refresh=False):
        return list_sensors(
            self, dict.items(self.sensor), filter, strategy, status,
            use_python_identifiers, tuple, refresh)

    @tornado.gen.coroutine
    def set_sampling_strategies(self, filter, strategy_and_parms):
        """Set a strategy for all sensors matching the filter, including unseen sensors
        The strategy should persist across sensor disconnect/reconnect.

        filter : str
            Filter for sensor names
        strategy_and_params : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.

        Returns
        -------
        done : tornado Future
            Resolves when done
        """
        sensor_list = yield self.list_sensors(filter=filter)
        sensor_dict = {}
        for sens in sensor_list:
            # Set the strategy on each sensor
            try:
                sensor_name = sens.object.normalised_name
                yield self.set_sampling_strategy(sensor_name, strategy_and_parms)
                sensor_dict[sensor_name] = strategy_and_parms
            except Exception as exc:
                self._logger.exception(
                    'Unhandled exception trying to set sensor strategies {!r} for {} ({})'
                    .format(strategy_and_parms, sens, exc))
                sensor_dict[sensor_name] = None
        # Otherwise, depend on self._add_sensors() to handle it from the cache when the sensor appears\
        raise tornado.gen.Return(sensor_dict)

    @tornado.gen.coroutine
    def set_sampling_strategy(self, sensor_name, strategy_and_parms):
        """Set a strategy for a sensor even if it is not yet known.
        The strategy should persist across sensor disconnect/reconnect.

        sensor_name : str
            Name of the sensor
        strategy_and_params : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.

        Returns
        -------
        done : tornado Future
            Resolves when done
        """
        sensor_name = resource.escape_name(sensor_name)
        sensor_obj = dict.get(self._sensor, sensor_name)
        self._sensor_strategy_cache[sensor_name] = strategy_and_parms
        sensor_dict = {}
        self._logger.debug(
                'Cached strategy {} for sensor {}'
                .format(strategy_and_parms, sensor_name))
        if sensor_obj:
            # The sensor exists, so set the strategy and continue. Log errors,
            # but don't raise anything
            try:
                yield sensor_obj.set_sampling_strategy(strategy_and_parms)
                sensor_dict[sensor_name] = strategy_and_parms
            except Exception as exc:
                self._logger.exception(
                    'Unhandled exception trying to set sensor strategy {!r} for sensor {} ({})'
                    .format(strategy_and_parms, sensor_name, exc))
                sensor_dict[sensor_name] = str(exc)
        # Otherwise, depend on self._add_sensors() to handle it from the cache when the sensor appears
        raise tornado.gen.Return(sensor_dict)

    def drop_sampling_strategy(self, sensor_name):
        """Drop the sampling strategy for the named sensor from the cache

        Calling :meth:`set_sampling_strategy` requires the requested strategy to
        be memorised so that it can automatically be reapplied.  This method
        causes the strategy to be forgotten.  There is no change to the current
        strategy.  No error is raised if there is no strategy to drop.

        Parameters
        ----------

        sensor_name : str
            Name of the sensor

        """
        sensor_name = resource.escape_name(sensor_name)
        # drop from both the internal cache, and the sensor manager's cache
        self._sensor_strategy_cache.pop(sensor_name, None)
        self._sensor_manager.drop_sampling_strategy(sensor_name)

    @tornado.gen.coroutine
    def set_sensor_listener(self, sensor_name, listener):
        """Set a sensor listener for a sensor even if it is not yet known
        The listener registration should persist across sensor disconnect/reconnect.

        sensor_name : str
            Name of the sensor
        listener : callable
            Listening callable that will be registered on the named sensor when it becomes
            available. Callable as for :meth:`KATCPSensor.register_listener`

        """

        sensor_name = resource.escape_name(sensor_name)
        sensor_obj = dict.get(self._sensor, sensor_name)
        self._sensor_listener_cache[sensor_name].append(listener)
        sensor_dict = {}
        self._logger.debug(
                'Cached listener {} for sensor {}'
                .format(listener, sensor_name))
        if sensor_obj:
            # The sensor exists, so register the listener and continue.
            try:
                sensor_obj.register_listener(listener, reading=True)
                sensor_dict[sensor_name] = listener
                self._logger.debug(
                    'Registered listener {} for sensor {}'
                    .format(listener, sensor_name))
            except Exception as exc:
                self._logger.exception(
                    'Unhandled exception trying to set sensor listener {} for sensor {} ({})'
                    .format(listener, sensor_name, exc))
                sensor_dict[sensor_name] = str(exc)
        # Otherwise, depend on self._add_sensors() to handle it from the cache when the sensor appears
        raise tornado.gen.Return(sensor_dict)

    def _request_factory(self, **request_description):
        return KATCPClientResourceRequest(
            request_description, self._inspecting_client, self.is_active)

    @tornado.gen.coroutine
    def _inspecting_client_state_callback(self, state, model_changes):
        self._logger.debug('{}: Received {}, {}'
                  .format(self.address_string, state, model_changes))
        if state.connected:
            if not state.synced:
                self._logger.debug('{}: Setting state to "syncing"'.format(
                    self.address_string))
                self._state.set_state('syncing')
                if model_changes:
                    self._logger.debug('{}: handling model updates: {}'.format(
                        self.address_string, model_changes))

                    yield self._update_model(model_changes)
                    self._logger.debug('{}: finished handling model updates'
                              .format(self.address_string))
                if state.data_synced:
                    # Reapply cached sensor strategies. Can only be done if
                    # data_synced==True, or else the
                    # self._inspecting_client.future_check_sensor() will deadlock
                    self._logger.debug('{}: Reapplying sampling strategies'
                              .format(self.address_string))
                    yield self._sensor_manager.reapply_sampling_strategies()
                    self._logger.debug('{}: Done Reapplying sampling strategies'
                              .format(self.address_string))
            else:
                self._logger.debug('{}: Setting state to "synced"'
                          .format(self.address_string))
                self._state.set_state('synced')
        else:
            self._logger.debug('{}: Setting state to "disconnected"'
                      .format(self.address_string))
            self._state.set_state('disconnected')

        self._logger.debug('Done with _inspecting_client_state_callback')

    @tornado.gen.coroutine
    def _update_model(self, model_changes):
        if 'requests' in model_changes:
            self._logger.debug('Removing requests')
            yield self._remove_requests(model_changes.requests.removed)
            self._logger.debug('Adding requests')
            yield self._add_requests(model_changes.requests.added)
            self._logger.debug('Done with requests')
        if 'sensors' in model_changes:
            self._logger.debug('Removing sensors')
            yield self._remove_sensors(model_changes.sensors.removed)
            self._logger.debug('Adding sensors')
            yield self._add_sensors(model_changes.sensors.added)
            self._logger.debug('Done with sensors')
        self._logger.debug('Done with model')


    @tornado.gen.coroutine
    def _add_requests(self, request_keys):
        # Instantiate KATCPRequest instances and store on self.req

        # Use LatencyTimer to avoid starving the ioloop
        latency_timer = LatencyTimer(self.MAX_LOOP_LATENCY)

        request_instance_fut = {}
        for key in request_keys:
            fut = request_instance_fut[key] = (
                self._inspecting_client.future_get_request(key))
            latency_timer.check_future(fut)
            if latency_timer.time_to_yield():
                yield tornado.gen.moment

        request_instances = yield request_instance_fut

        added_names = []
        for r_name, r_obj in request_instances.items():
            r_name_escaped = resource.escape_name(r_name)
            if r_name_escaped in self.always_excluded_requests:
                continue
            if self.controlled or r_name_escaped in self.always_allowed_requests:
                self._req[r_name_escaped] = r_obj
                added_names.append(r_name_escaped)

        if self.parent and added_names:
            self.parent._child_add_requests(self, added_names)

    @tornado.gen.coroutine
    def _remove_requests(self, request_keys):
        # Remove KATCPRequest instances from self.req
        removed_names = []
        for r_name in request_keys:
            r_name_escaped = resource.escape_name(r_name)
            # Must not raise exception when popping a non-existing request, since it may
            # never have been added due to request exclusion rules.
            if dict.pop(self.req, r_name_escaped, None):
                removed_names.append(r_name_escaped)

        if self.parent and removed_names:
            self.parent._child_remove_requests(self, removed_names)

    @tornado.gen.coroutine
    def _add_sensors(self, sensor_keys):
        # Use LatencyTimer to avoid starving the ioloop
        latency_timer = LatencyTimer(self.MAX_LOOP_LATENCY)
        sensor_instance_fut = {}
        # Get KATCPSensor instance futures from inspecting client
        for key in sensor_keys:
            fut = sensor_instance_fut[key] = (
                self._inspecting_client.future_get_sensor(key))
            latency_timer.check_future(fut)
            if latency_timer.time_to_yield():
                yield tornado.gen.moment

        sensor_instances = yield sensor_instance_fut
        # Store KATCPSensor instances in self.sensor
        added_names = []
        for s_name, s_obj in list(sensor_instances.items()):
            s_name_escaped = resource.escape_name(s_name)
            self._sensor[s_name_escaped] = s_obj
            preset_strategy = self._sensor_strategy_cache.get(s_name_escaped)
            if preset_strategy:
                self._logger.debug('Setting preset strategy for sensor {} to {!r}'
                                   .format(s_name, preset_strategy))
                try:
                    yield s_obj.set_sampling_strategy(preset_strategy)
                except Exception:
                    self._logger.exception(
                        'Exception trying to pre-set sensor strategy for sensor {}'
                        .format(s_name))
            preset_listeners = self._sensor_listener_cache.get(s_name_escaped)
            if preset_listeners:
                try:
                    for listener in preset_listeners:
                        s_obj.register_listener(listener, reading=True)
                except Exception:
                    self._logger.exception(
                        'Exception trying to pre-set sensor listeners for sensor {}'
                        .format(s_name))

            added_names.append(s_name_escaped)

        if self.parent:
            self.parent._child_add_sensors(self, added_names)

    @tornado.gen.coroutine
    def _remove_sensors(self, sensor_keys):
        # Remove KATCPSensor instances from self.sensor
        removed_names = []
        for s_name in sensor_keys:
            s_name_escaped = resource.escape_name(s_name)
            if s_name_escaped in self.sensor:
                dict.pop(self.sensor, s_name_escaped)
                removed_names.append(s_name_escaped)

        if self.parent:
            self.parent._child_remove_sensors(self, removed_names)

    def stop(self):
        self._inspecting_client.stop()

    def until_stopped(self, timeout=None):
        """Return future that resolves when the inspecting client has stopped

        See the `DeviceClient.until_stopped` docstring for parameter
        definitions and more info.
        """
        return self._inspecting_client.until_stopped(timeout)

    def __repr__(self):
        return '<{module}.{classname}(name={name}) at 0x{id:x}>'.format(
            module=self.__class__.__module__,
            classname=self.__class__.__name__,
            name=self.name, id=id(self))


class KATCPClientResourceSensorsManager(object):
    """Implementation of KATSensorsManager ABC for a directly-connected client

    Assumes that all methods are called from the same ioloop context
    """

    def __init__(self, inspecting_client, resource_name, logger=log):
        self._inspecting_client = inspecting_client
        self.time = inspecting_client.ioloop.time
        self._strategy_cache = {}
        self._resource_name = resource_name
        self._logger = logger

    def _get_strategy_cache_key(self, sensor_name):
        """Lookup sensor name in cache, allowing names in escaped form

        The strategy cache uses the normal KATCP sensor names as the keys.
        In order to allow access using an escaped sensor name, this method
        tries to find the normal form of the name.

        Returns
        -------

        key : str
            If there is a match, the cache key is returned.
            If no match, then the sensor_name is returned unchanged.

        """
        # try for a direct match first, otherwise do full comparison
        if sensor_name in self._strategy_cache:
            return sensor_name
        else:
            escaped_name = resource.escape_name(sensor_name)
            for key in self._strategy_cache:
                escaped_key = resource.escape_name(key)
                if escaped_key == escaped_name:
                    return key
        # no match
        return sensor_name

    @property
    def resource_name(self):
        return self._resource_name

    def sensor_factory(self, **sensor_description):
        # kwargs as for inspecting_client.InspectingClientAsync.sensor_factory
        sens = resource.KATCPSensor(sensor_description, self)
        sensor_name = sensor_description['name']
        cached_strategy = self._strategy_cache.get(sensor_name)
        if cached_strategy:
            log_future_exceptions(self._logger, self.set_sampling_strategy(
                sensor_name, cached_strategy))
        return sens

    def get_sampling_strategy(self, sensor_name):
        """Get the current sampling strategy for the named sensor

        Parameters
        ----------

        sensor_name : str
            Name of the sensor (normal or escaped form)

        Returns
        -------

        strategy : tuple of str
            contains (<strat_name>, [<strat_parm1>, ...]) where the strategy names and
            parameters are as defined by the KATCP spec
        """
        cache_key = self._get_strategy_cache_key(sensor_name)
        cached = self._strategy_cache.get(cache_key)
        if not cached:
            return resource.normalize_strategy_parameters('none')
        else:
            return cached

    @tornado.gen.coroutine
    def set_sampling_strategy(self, sensor_name, strategy_and_params):
        """Set the sampling strategy for the named sensor

        Parameters
        ----------

        sensor_name : str
            Name of the sensor
        strategy_and_params : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the
            strategy names and parameters are as defined by the KATCP spec. As
            str contains the same elements in space-separated form.

        Returns
        -------
        sensor_strategy : tuple
            (success, info) with

            success : bool
                True if setting succeeded for this sensor, else False
            info : tuple
               Normalised sensor strategy and parameters as tuple if
               success == True else, sys.exc_info() tuple for the error
               that occurred.
        """
        try:
            strategy_and_params = resource.normalize_strategy_parameters(
                strategy_and_params)
            self._strategy_cache[sensor_name] = strategy_and_params
            reply = yield self._inspecting_client.wrapped_request(
                'sensor-sampling', sensor_name, *strategy_and_params)
            if not reply.succeeded:
                raise KATCPSensorError('Error setting strategy for sensor {0}: \n'
                                       '{1!s}'.format(sensor_name, reply))
            sensor_strategy = (True, strategy_and_params)
        except Exception as e:
            self._logger.exception('Exception found!')
            sensor_strategy = (False, str(e))
        raise tornado.gen.Return(sensor_strategy)

    def drop_sampling_strategy(self, sensor_name):
        """Drop the sampling strategy for the named sensor from the cache

        Calling :meth:`set_sampling_strategy` requires the sensor manager to
        memorise the requested strategy so that it can automatically be reapplied.
        If the client is no longer interested in the sensor, or knows the sensor
        may be removed from the server, then it can use this method to ensure the
        manager forgets about the strategy.  This method will not change the current
        strategy.  No error is raised if there is no strategy to drop.

        Parameters
        ----------

        sensor_name : str
            Name of the sensor (normal or escaped form)

        """
        cache_key = self._get_strategy_cache_key(sensor_name)
        self._strategy_cache.pop(cache_key, None)

    @tornado.gen.coroutine
    def reapply_sampling_strategies(self):
        """Reapply all sensor strategies using cached values"""
        check_sensor = self._inspecting_client.future_check_sensor
        for sensor_name, strategy in list(self._strategy_cache.items()):
            try:
                sensor_exists = yield check_sensor(sensor_name)
                if not sensor_exists:
                    self._logger.warn('Did not set strategy for non-existing sensor {}'
                                      .format(sensor_name))
                    continue

                result = yield self.set_sampling_strategy(sensor_name, strategy)
            except KATCPSensorError as e:
                self._logger.error('Error reapplying strategy for sensor {0}: {1!s}'
                                   .format(sensor_name, e))
            except Exception:
                self._logger.exception('Unhandled exception reapplying strategy for '
                                       'sensor {}'.format(sensor_name), exc_info=True)

    @tornado.gen.coroutine
    @steal_docstring_from(resource.KATCPSensorsManager.poll_sensor)
    def poll_sensor(self, sensor_name):
        reply = yield self._inspecting_client.wrapped_request(
            'sensor-value', sensor_name)
        if not reply.succeeded:
            raise KATCPSensorError('Error polling sensor {0}: \n'
                                   '{1!s}'.format(sensor_name, reply))
# Register with the ABC
resource.KATCPSensorsManager.register(KATCPClientResourceSensorsManager)

class KATCPClientResourceRequest(resource.KATCPRequest):
    """Callable wrapper around a KATCP request

    """
    def __init__(self, request_description, client, is_active=lambda : True):
        """Initialize request with given description and network client

        Parameters
        ----------
        request_description : dict
            Description of KATCP request as required by
            :class:`katcp.resource.KATCPRequest` init method
        client : client obj
            KATCP client connected to the KATCP resource that exposes a wrapped_request()
            method like :meth:`ReplyWrappedInspectingClientAsync.wrapped_request`.
        is_active : callable, optional
            Returns True if this request is active, else False

        """
        self._client = client
        super(KATCPClientResourceRequest, self).__init__(request_description, is_active)

    def issue_request(self, *args, **kwargs):
        """Issue the wrapped request to the server.

        Parameters
        ----------
        *args : list of objects
            Arguments to pass on to the request.

        Keyword Arguments
        -----------------
        timeout : float or None, optional
            Timeout after this amount of seconds (keyword argument).
        mid : None or int, optional
            Message identifier to use for the request message. If None, use either
            auto-incrementing value or no mid depending on the KATCP protocol version
            (mid's were only introduced with KATCP v5) and the value of the `use_mid`
            argument. Defaults to None.
        use_mid : bool
            Use a mid for the request if True.

        Returns
        -------
        future object that resolves with an :class:`katcp.resource.KATCPReply`
        instance

        """
        timeout = kwargs.pop('timeout', None)
        if timeout is None:
            timeout = self.timeout_hint
        kwargs['timeout'] = timeout
        return self._client.wrapped_request(self.name, *args, **kwargs)

class GroupRequest(object):
    """Couroutine wrapper around a specific KATCP request for a group of clients.

    Each available KATCP request supported by group has an associated
    :class:`GroupRequest` object in the hierarchy. This wrapper is mainly for
    interactive convenience. It provides the KATCP request help string as a
    docstring accessible via IPython's question mark operator.

    Parameters
    ----------

    Call parameters are all forwarded to the :class:`KATCPRequest` instance of each
    client in the group.

    Returns
    -------
    Returns a tornado future that resolves with a :class:`GroupResults` instance that
    contains the replies of each client. If a particular client does not have the request,
    its result is None.

    """
    def __init__(self, group, name, description):
        """Initialise the GroupRequest

        Parameters
        ----------
        group : :class:`ClientGroup` object
            Client group to which requests will be sent
        name : string
            Name of the KATCP request
        description : string
            Help string associated with this KATCP request

        """

        self.group = group
        self.name = name
        self.description = description
        self.__doc__ = '\n'.join(('KATCP Documentation',
                                  '===================',
                                  description,
                                  'GroupRequest Documentation',
                                  '==========================',
                                  self.__doc__ or ''))

    @tornado.gen.coroutine
    def __call__(self, *args, **kwargs):
        result_futures = {}
        none_future = Future()
        none_future.set_result(None)
        for client in self.group.clients:
            request_method = getattr(client.req, self.name, None)
            if request_method:
                result_futures[client.name] = request_method(*args, **kwargs)
            else:
                result_futures[client.name] = none_future

        results = yield result_futures
        raise Return(GroupResults(results))


class GroupResults(dict):
    """The result of a group request.

    This has a dictionary interface, with the client names as keys and the
    corresponding replies from each client as values. The replies are stored as
    :class:`KATCPReply` objects, or are None for clients
    that did not support the request.

    The result will evaluate to a truthy value if all the requests succeeded, i.e.
    ::

        if result:
            handle_success()
        else:
            handle_failure()

    should work as expected.

    """
    def __bool__(self):
        """True if katcp request succeeded on all clients."""
        return all(self.values())

    if PY2:
        __nonzero__ = __bool__

    @property
    def succeeded(self):
        """True if katcp request succeeded on all clients."""
        return bool(self)


class ClientGroup(object):
    """Create a group of similar clients.

    Parameters
    ----------
    name : str
        Name of the group of clients.
    clients : list of :class:`KATCPResource` objects
        Clients to put into the group.
    """
    def __init__(self, name, clients):
        self.name = name
        self._clients_dirty = True
        self.clients = tuple(clients)

    def __iter__(self):
        """Iterate over client members of group."""
        return iter(self.clients)

    def __getitem__(self, index):
        """Get the client at specific index of group."""
        return self.clients[index]

    def __len__(self):
        """Number of client members in group."""
        return len(self.clients)

    @property
    def req(self):
        if self._clients_dirty:
            if any(client.dummy_unknown_requests for client in self.clients):
                DummyRequest = partial(
                    resource.KATCPDummyRequest,
                    {'name':'dummy', 'description': 'No help for dummies',
                     'timeout_hint': None})
                self._req = DefaultAttrDict(DummyRequest)
            else:
                self._req = AttrDict()

            for client in self.clients:
                for name, request in dict.items(client.req):
                    if name not in self._req:
                        self._req[name] = GroupRequest(self, name,
                                                       request.description)
            self._clients_dirty = False

        return self._req

    def client_updated(self, client):
        """Called to notify this group that a client has been updated."""
        assert client in self.clients
        self._clients_dirty = True

    def is_connected(self):
        """Indication of the connection state of all clients in the group"""
        return all([c.is_connected() for c in self.clients])

    @tornado.gen.coroutine
    def set_sampling_strategies(self, filter, strategy_and_params):
        """Set sampling strategy for the sensors of all the group's clients.

        Only sensors that match the specified filter are considered. See the
        `KATCPResource.set_sampling_strategies` docstring for parameter
        definitions and more info.

        Returns
        -------
        sensors_strategies : tornado Future
           Resolves with a dict with client names as keys and with the value as
           another dict. The value dict is similar to the return value
           described in the `KATCPResource.set_sampling_strategies` docstring.
        """
        futures_dict = {}
        for res_obj in self.clients:
            futures_dict[res_obj.name] = res_obj.set_sampling_strategies(
                filter, strategy_and_params)
        sensors_strategies = yield futures_dict
        raise tornado.gen.Return(sensors_strategies)

    @tornado.gen.coroutine
    def set_sampling_strategy(self, sensor_name, strategy_and_params):
        """Set sampling strategy for the sensors of all the group's clients.

        Only sensors that match the specified filter are considered. See the
        `KATCPResource.set_sampling_strategies` docstring for parameter
        definitions and more info.

        Returns
        -------
        sensors_strategies : tornado Future
           Resolves with a dict with client names as keys and with the value as
           another dict. The value dict is similar to the return value
           described in the `KATCPResource.set_sampling_strategies` docstring.
        """
        futures_dict = {}
        for res_obj in self.clients:
            futures_dict[res_obj.name] = res_obj.set_sampling_strategy(
                sensor_name, strategy_and_params)
        sensors_strategies = yield futures_dict
        raise tornado.gen.Return(sensors_strategies)

    @tornado.gen.coroutine
    def wait(self, sensor_name, condition_or_value, timeout=5.0, quorum=None,
             max_grace_period=1.0):
        """Wait for sensor present on all group clients to satisfy a condition.

        Parameters
        ----------
        sensor_name : string
            The name of the sensor to check
        condition_or_value : obj or callable, or seq of objs or callables
            If obj, sensor.value is compared with obj. If callable,
            condition_or_value(reading) is called, and must return True if its
            condition is satisfied. Since the reading is passed in, the value,
            status, timestamp or received_timestamp attributes can all be used
            in the check.
        timeout : float or None
            The total timeout in seconds (None means wait forever)
        quorum : None or int or float
            The number of clients that are required to satisfy the condition,
            as either an explicit integer or a float between 0 and 1 indicating
            a fraction of the total number of clients, rounded up. If None,
            this means that all clients are required (the default). Be warned
            that a value of 1.0 (float) indicates all clients while a value
            of 1 (int) indicates a single client...
        max_grace_period : float or None
            After a quorum or initial timeout is reached, wait up to this long
            in an attempt to get the rest of the clients to satisfy condition
            as well (achieving effectively a full quorum if all clients behave)

        Returns
        -------
        This command returns a tornado Future that resolves with True when a
        quorum of clients satisfy the sensor condition, or False if a quorum
        is not reached after a given timeout period (including a grace period).

        Raises
        ------
        :class:`KATCPSensorError`
            If any of the sensors do not have a strategy set, or if the named
            sensor is not present

        """
        if quorum is None:
            quorum = len(self.clients)
        elif quorum > 1:
            if not isinstance(quorum, int):
                raise TypeError('Quorum parameter %r must be an integer '
                                'if outside range [0, 1]' % (quorum,))
        elif isinstance(quorum, float):
            quorum = int(math.ceil(quorum * len(self.clients)))
        if timeout and max_grace_period:
            # Avoid having a grace period longer than or equal to timeout
            grace_period = min(max_grace_period, timeout / 2.)
            initial_timeout = timeout - grace_period
        else:
            grace_period = max_grace_period
            initial_timeout = timeout
        # Build dict of futures instead of list as this will be easier to debug
        futures = {}
        for client in self.clients:
            f = client.wait(sensor_name, condition_or_value, initial_timeout)
            futures[client.name] = f
        # No timeout required here as all futures will resolve after timeout
        initial_results = yield until_some(done_at_least=quorum, **futures)
        results = dict(initial_results)
        # Identify stragglers and let them all respond within grace period
        stragglers = {}
        for client in self.clients:
            if not results.get(client.name, False):
                f = client.wait(sensor_name, condition_or_value, grace_period)
                stragglers[client.name] = f
        rest_of_results = yield until_some(**stragglers)
        results.update(dict(rest_of_results))

        class TestableDict(dict):
            """Dictionary of results that can be tested for overall success."""
            def __bool__(self):
                return sum(self.values()) >= quorum

            if PY2:
                __nonzero__ = __bool__

        raise tornado.gen.Return(TestableDict(results))


class KATCPClientResourceContainer(resource.KATCPResource):
    """Class for containing multiple :class:`KATCPClientResource` instances.

    Provides aggregate `sensor` and `req` attributes containing the union of all the
    sensors in requests in the contained resources. Names are prefixed with <resname>_,
    where <resname> is the name of the resource to which the sensor / request belongs
    except for aggregate sensors that starts with ``agg_``.

    """
    @property
    def req(self):
        if self._children_dirty:
            self._req = self._create_attrdict_from_children('req')
            self._children_dirty = False

        return self._req

    @property
    def sensor(self):
        if self._children_dirty:
            self._sensor = self._create_attrdict_from_children('sensor')
            self._children_dirty = False

        return self._sensor

    @property
    def sensors(self):
        return self.sensor

    @property
    def address(self):
        return None

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def parent(self):
        return None

    @property
    def children(self):
        return self._children

    @property
    def groups(self):
        return self._groups

    def __init__(self, resources_spec, logger=log):
        """Initialise Container with specifications for all the child resources

        Parameters
        ----------

        resources_spec : dict containing the specs of the contained resources. Keys:
          "name" : str, name of this collection of resources
          "description : str (optional), description of this collection of resources
          "clients" : dict, with keys:
            <name> : resource specifications for :class:`KATCPClientResource` where <name>
                     is the name of the resource. Note that the `name` key in the
                     individual resource spec dicts should not be specified, and will
                     automatically be filled using the <name> key above.
          "groups" : optional dict of resource groupings with keys: <group-name> : seq of
            <group-member>* where <group-name> is the name of the group and
            <group-member>* is a subset of clients' names as in the "clients" dict
            above. Also, the <group-name> set must be disjoint from the <name> set above.

        logger : object, optional
           Python Logger object to log to. Default is the module logger
        """
        super(KATCPClientResourceContainer, self).__init__()
        self._resources_spec = resources_spec
        self._logger = logger
        self._name = resources_spec['name']
        self._description = resources_spec.get('description', '')
        self._children_dirty = True   # Are we out of sync with the children?
        self._init_resources()
        self._init_groups()
        self.set_ioloop()

    def _init_resources(self):
        resources = self._resources_spec['clients']
        children = AttrDict()
        for res_name, res_spec in list(resources.items()):
            # Make a copy since we'll be modifying the dict
            res_spec = dict(res_spec)
            res_spec['name'] = res_name
            res = self.client_resource_factory(
                res_spec, parent=self, logger=self._logger)
            children[resource.escape_name(res_name)] = res
        self._children = children

    def client_resource_factory(self, res_spec, parent, logger):
        """Return an instance of :class:`KATCPClientResource` or similar

        Provided to ease testing. Overriding this method allows deep brain surgery. See
        :func:`katcp.fake_clients.fake_KATCP_client_resource_factory`

        """
        return KATCPClientResource(res_spec, parent=self, logger=logger)

    def _init_groups(self):
        group_configs = self._resources_spec.get('groups', {})
        groups = AttrDict()
        for group_name, group_client_names in list(group_configs.items()):
            group_clients = tuple(self.children[resource.escape_name(cn)]
                                  for cn in group_client_names)
            group = ClientGroup(group_name, group_clients)
            groups[resource.escape_name(group_name)] = group

        self._groups = groups

    def add_group(self, group_name, group_client_names):
        """Add a new :class:`ClientGroup` to container groups member.

        Add the group named *group_name* with sequence of client names to the
        container groups member. From there it will be wrapped appropriately
        in the higher-level thread-safe container.
        """
        group_configs = self._resources_spec.get('groups', {})
        group_configs[group_name] = group_client_names
        self._resources_spec['groups'] = group_configs
        self._init_groups()

    def set_ioloop(self, ioloop=None):
        """Set the tornado ioloop to use

        Defaults to tornado.ioloop.IOLoop.current() if set_ioloop() is not called or if
        ioloop=None. Must be called before start()
        """
        ioloop = ioloop or tornado.ioloop.IOLoop.current()
        self.ioloop = ioloop
        for res in dict.values(self.children):
            res.set_ioloop(ioloop)

    def is_connected(self):
        """Indication of the connection state of all children"""
        return all([r.is_connected() for r in dict.values(self.children)])

    def start(self):
        """Start and connect all the subordinate clients"""
        for res in dict.values(self.children):
            res.start()

    @tornado.gen.coroutine
    def until_synced(self, timeout=None):
        """Return a tornado Future; resolves when all subordinate clients are synced"""
        futures = [r.until_synced(timeout) for r in dict.values(self.children)]
        yield tornado.gen.multi(futures, quiet_exceptions=tornado.gen.TimeoutError)

    @tornado.gen.coroutine
    def until_not_synced(self, timeout=None):
        """Return a tornado Future; resolves when any subordinate client is not synced"""
        yield until_any(*[r.until_not_synced() for r in dict.values(self.children)],
                        timeout=timeout)

    def until_any_child_in_state(self, state, timeout=None):
        """Return a tornado Future; resolves when any client is in specified state"""
        return until_any(*[r.until_state(state) for r in dict.values(self.children)],
                         timeout=timeout)

    @tornado.gen.coroutine
    def until_all_children_in_state(self, state, timeout=None):
        """Return a tornado Future; resolves when all clients are in specified state"""
        futures = [r.until_state(state, timeout=timeout)
                   for r in dict.values(self.children)]
        yield tornado.gen.multi(futures, quiet_exceptions=tornado.gen.TimeoutError)

    @steal_docstring_from(resource.KATCPResource.list_sensors)
    def list_sensors(self, filter="", strategy=False, status="",
                     use_python_identifiers=True, tuple=False, refresh=False):
        return list_sensors(
            self, dict.items(self.sensor), filter, strategy, status,
            use_python_identifiers, tuple, refresh)

    @tornado.gen.coroutine
    def set_sampling_strategies(self, filter, strategy_and_parms):
        """Set sampling strategies for filtered sensors - these sensors have to exists."""
        result_list = yield self.list_sensors(filter=filter)
        sensor_dict = {}
        for result in result_list:
            sensor_name = result.object.normalised_name
            resource_name = result.object.parent_name
            if resource_name not in sensor_dict:
                sensor_dict[resource_name] = {}
            try:
                resource_obj = self.children[resource_name]
                yield resource_obj.set_sampling_strategy(sensor_name, strategy_and_parms)
                sensor_dict[resource_name][sensor_name] = strategy_and_parms
                self._logger.debug(
                    'Set sampling strategy on resource %s for %s'
                    % (resource_name, sensor_name))
            except Exception as exc:
                self._logger.error(
                    'Cannot set sampling strategy on resource %s for %s (%s)'
                    % (resource_name, sensor_name, exc))
                sensor_dict[resource_name][sensor_name] = None
        raise tornado.gen.Return(sensor_dict)

    @tornado.gen.coroutine
    def set_sampling_strategy(self, sensor_name, strategy_and_parms):
        """Set sampling strategies for the specific sensor - this sensor has to exist"""
        result_list = yield self.list_sensors(filter="^"+sensor_name+"$") #exact match
        sensor_dict = {}
        for result in result_list:
            sensor_name = result.object.normalised_name
            resource_name = result.object.parent_name
            if resource_name not in sensor_dict:
                sensor_dict[resource_name] = {}
            try:
                resource_obj = self.children[resource_name]
                yield resource_obj.set_sampling_strategy(sensor_name, strategy_and_parms)
                sensor_dict[resource_name][sensor_name] = strategy_and_parms
                self._logger.debug(
                    'Set sampling strategy on resource %s for %s'
                    % (resource_name, sensor_name))
            except Exception as exc:
                self._logger.error(
                    'Cannot set sampling strategy on resource %s for %s (%s)'
                    % (resource_name, sensor_name, exc))
                sensor_dict[resource_name][sensor_name] = None
        raise tornado.gen.Return(sensor_dict)

    @tornado.gen.coroutine
    def set_sensor_listener(self, sensor_name, listener):
        """Set listener for the specific sensor - this sensor has to exists."""
        result_list = yield self.list_sensors(filter="^"+sensor_name+"$") #exact match
        sensor_dict = {}
        for result in result_list:
            sensor_name = result.object.normalised_name
            resource_name = result.object.parent_name
            if resource_name not in sensor_dict:
                sensor_dict[resource_name] = {}
            try:
                resource_obj = self.children[resource_name]
                yield resource_obj.set_sensor_listener(sensor_name, listener)
                sensor_dict[resource_name][sensor_name] = listener
                self._logger.debug(
                    'Set sensor listener on resource %s for %s'
                    % (resource_name, sensor_name))
            except Exception as exc:
                self._logger.error(
                    'Cannot set sensor listener on resource %s for %s (%s)'
                    % (resource_name, sensor_name, exc))
                sensor_dict[resource_name][sensor_name] = None
        raise tornado.gen.Return(sensor_dict)

    def add_child_resource_client(self, res_name, res_spec):
        """Add a resource client to the container and start the resource connection"""
        res_spec = dict(res_spec)
        res_spec['name'] = res_name
        res = self.client_resource_factory(
                res_spec, parent=self, logger=self._logger)
        self.children[resource.escape_name(res_name)] = res;
        self._children_dirty = True
        res.set_ioloop(self.ioloop)
        res.start()
        return res

    def _create_attrdict_from_children(self, attr):
        attrdict = AttrDict()
        for child_name, child_resource in dict.items(self.children):
            prefix = resource.escape_name(child_name) + '_'
            for item_name, item in dict.items(getattr(child_resource, attr)):
                # Do not prefix aggregate sensors with "parent_name_"
                if item_name.startswith("agg_"):
                    full_item_name = item_name
                else:
                    full_item_name = prefix + item_name
                attrdict[full_item_name] = item
        return attrdict

    def stop(self):
        """Stop all child resources"""
        for child_name, child in dict.items(self.children):
            # Catch child exceptions when stopping so we make sure to stop all children
            # that want to listen.
            try:
                child.stop()
            except Exception:
                self._logger.exception('Exception stopping child {!r}'
                                       .format(child_name))

    def until_stopped(self, timeout=None):
        """Return dict of futures that resolve when each child resource has stopped

        See the `DeviceClient.until_stopped` docstring for parameter
        definitions and more info.
        """
        futures = {}
        for child_name, child in dict.items(self.children):
            futures[child_name] = child.until_stopped(timeout)
        return futures

    @steal_docstring_from(resource.KATCPResource.wait)
    def wait(self, sensor_name, condition_or_value, timeout=5):
        raise NotImplementedError

    def _child_add_requests(self, child, sensor_keys):
        assert resource.escape_name(child.name) in self.children
        self._children_dirty = True
        self._dirty_groups(child)

    def _child_remove_requests(self, child, sensor_keys):
        assert resource.escape_name(child.name) in self.children
        self._children_dirty = True
        self._dirty_groups(child)

    def _child_add_sensors(self, child, sensor_keys):
        assert resource.escape_name(child.name) in self.children
        self._children_dirty = True

    def _child_remove_sensors(self, child, sensor_keys):
        assert resource.escape_name(child.name) in self.children
        self._children_dirty = True

    def _dirty_groups(self, child):
        groups_spec = self._resources_spec.get('groups', {})
        for group_name, group in dict.items(self.groups):
            if child.name in groups_spec[group_name]:
                group.client_updated(child)

    def __repr__(self):
        return '<{module}.{classname}(name={name}) at 0x{id:x}>'.format(
            module=self.__class__.__module__,
            classname=self.__class__.__name__,
            name=self.name, id=id(self))


class ThreadSafeKATCPSensorWrapper(ThreadSafeMethodAttrWrapper):

    @property
    def sampling_strategy(self):
        return self._getattr('sampling_strategy')


class ThreadSafeKATCPClientResourceRequestWrapper(ThreadSafeMethodAttrWrapper):
    @property
    def __call__(self):
        return self._ioloop_wrapper.decorate_callable(self.__subject__.__call__)


class MappingProxy(collections.Mapping):

    def __init__(self, mapping, wrapper):
        self._mapping = mapping
        self._wrapper = wrapper

    def __iter__(self):
        return iter(self._mapping)

    def __len__(self):
        return len(self._mapping)

    def __contains__(self, x):
        return x in self._mapping

    def __getitem__(self, key):
        return self._wrapper(self._mapping[key])


class AttrMappingProxy(MappingProxy):

    def __getattr__(self, attr):
        return self._wrapper(getattr(self._mapping, attr))

    def __dir__(self):
        return list(self.keys())


class ThreadSafeKATCPClientGroupWrapper(ThreadSafeMethodAttrWrapper):
    """Thread safe wrapper for :class:`ClientGroup`"""

    __slots__ = ['RequestWrapper', 'ResourceWrapper']

    def __init__(self, subject, ioloop_wrapper):
        self.RequestWrapper = partial(ThreadSafeKATCPClientResourceRequestWrapper,
                                      ioloop_wrapper=ioloop_wrapper)
        self.ResourceWrapper = partial(ThreadSafeKATCPClientResourceWrapper,
                                       ioloop_wrapper=ioloop_wrapper)
        super(ThreadSafeKATCPClientGroupWrapper, self).__init__(
            subject, ioloop_wrapper)

    def __iter__(self):
        """Iterate over client members of group."""
        return iter(self.clients)

    def __getitem__(self, index):
        """Get the client at specific index of group."""
        return self.clients[index]

    @property
    def req(self):
        return AttrMappingProxy(self.__subject__.req, self.RequestWrapper)

    @property
    def clients(self):
        async_clients = self.__subject__.clients
        blocking_clients = [self.ResourceWrapper(ac) for ac in async_clients]
        return blocking_clients


class ThreadSafeKATCPClientResourceWrapper(ThreadSafeMethodAttrWrapper):
    """Should work with both KATCPClientResource or KATCPClientResourceContainer"""

    __slots__ = ['ResourceWrapper', 'SensorWrapper', 'RequestWrapper',
                 'GroupWrapper']

    def __init__(self, subject, ioloop_wrapper):
        self.ResourceWrapper = partial(ThreadSafeKATCPClientResourceWrapper,
                                       ioloop_wrapper=ioloop_wrapper)
        self.SensorWrapper = partial(ThreadSafeKATCPSensorWrapper,
                                     ioloop_wrapper=ioloop_wrapper)
        self.RequestWrapper = partial(ThreadSafeKATCPClientResourceRequestWrapper,
                                      ioloop_wrapper=ioloop_wrapper)
        self.GroupWrapper = partial(ThreadSafeKATCPClientGroupWrapper,
                                       ioloop_wrapper=ioloop_wrapper)
        super(ThreadSafeKATCPClientResourceWrapper, self).__init__(
            subject, ioloop_wrapper)

    def __getattr__(self, attr):
        val = super(ThreadSafeKATCPClientResourceWrapper, self).__getattr__(attr)
        if isinstance(val, resource.KATCPResource):
            return self.ResourceWrapper(val)
        elif isinstance(val, ClientGroup):
            return self.GroupWrapper(val)
        else:
            return val

    @property
    def sensor(self):
        return AttrMappingProxy(self.__subject__.sensor, self.SensorWrapper)

    @property
    def sensors(self):
        return self.sensor

    @property
    def req(self):
        return AttrMappingProxy(self.__subject__.req, self.RequestWrapper)

    @property
    def groups(self):
        return AttrMappingProxy(self.__subject__.groups, self.GroupWrapper)

    @property
    def children(self):
        if self.__subject__.children:
            return AttrMappingProxy(self.__subject__.children, self.ResourceWrapper)
        else:
            return AttrDict()

    @property
    def parent(self):
        if self.__subject__.parent:
            return self.ResourceWrapper(self.__subject__.parent)


@tornado.gen.coroutine
def monitor_resource_sync_state(resource, callback, exit_event=None):
    """Coroutine that monitors a KATCPResource's sync state.

    Calls callback(True/False) whenever the resource becomes synced or unsynced. Will
    always do an initial callback(False) call. Exits without calling callback() if
    exit_event is set.

    Warning:  set the monitor's exit_event before stopping the resources being
    monitored, otherwise it could result in a memory leak.  The `until_synced()` or
    `until_not_synced()` methods could keep a reference to the resource alive.
    """
    exit_event = exit_event or AsyncEvent()
    callback(False)        # Initial condition, assume resource is not connected

    while not exit_event.is_set():
        # Wait for resource to be synced
        yield until_any(resource.until_synced(), exit_event.until_set())
        if exit_event.is_set():
            break       # If exit event is set we stop without calling callback
        else:
            callback(True)

        # Wait for resource to be un-synced
        yield until_any(resource.until_not_synced(), exit_event.until_set())
        if exit_event.is_set():
            break       # If exit event is set we stop without calling callback
        else:
            callback(False)



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

from tornado.concurrent import Future as tornado_Future
from tornado.util import ObjectDict

from katcp import Sensor
from katcp.client import *
from katcp.testutils import DeviceTestServer

logging.basicConfig(
    format="%(asctime)s %(name)s %(levelname)s %(funcName)s(%(filename)s:%(lineno)d)%(message)s",
    level=logging.DEBUG
)

def cb(*args):
    print(args)

try:
    d = DeviceTestServer('', 0)
    d.start(timeout=1)
    logging.info('Server started at port {0}'.format(d.bind_address[1]))
    c = AsyncClient('127.0.0.1', d.bind_address[1])
    c.enable_thread_safety()
    c.start(timeout=1)

#     rm = Message.request

#     #time.sleep(10000000)
    import IPython ; IPython.embed()
finally:
    d.stop()
    c.stop()



# sampling.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Strategies for sampling sensor values."""

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import logging
import os

from builtins import range
from functools import wraps

import tornado.ioloop

from _thread import get_ident as get_thread_ident
from future.utils import PY2

from .compat import is_bytes, is_text
from .core import Message, Sensor

log = logging.getLogger("katcp.sampling")

AGE_OF_UNIVERSE = 4.32329886e17  # approximate age of the universe


# pylint: disable-msg=W0142

def format_inform_v4(sensor, *reading):
    timestamp, status, value = sensor.format_reading(reading, 4)
    return Message.inform(
        "sensor-status", timestamp, "1", sensor.name, status, value)


def format_inform_v5(sensor, *reading):
    timestamp, status, value = sensor.format_reading(reading, 5)
    return Message.inform(
        "sensor-status", timestamp, "1", sensor.name, status, value)


def update_in_ioloop(update):
    """Decorator that ensures an update() method is run in the tornado ioloop.

    Does this by checking the thread identity. Requires that the object to
    which the method is bound has the attributes :attr:`_ioloop_thread_id`
    (the result of thread.get_ident() in the ioloop thread) and :attr:`ioloop`
    (the ioloop instance in use). Also assumes the signature
    `update(self, sensor, reading)` for the method.


    """
    @wraps(update)
    def wrapped_update(self, sensor, reading):
        if get_thread_ident() == self._ioloop_thread_id:
            update(self, sensor, reading)
        else:
            self.ioloop.add_callback(update, self, sensor, reading)

    return wrapped_update


class SampleStrategy(object):
    """Base class for strategies for sampling sensors.

    Parameters
    ----------
    inform_callback : callable, signature inform_callback(sensor_obj, reading)
        Callback to receive inform messages.
    sensor : Sensor object
        Sensor to sample.
    params : list of objects
        Custom sampling parameters.

    """

    # Sampling strategy constants
    NONE, AUTO, PERIOD, EVENT, DIFFERENTIAL, EVENT_RATE, DIFFERENTIAL_RATE = range(7)

    ## @brief Mapping from strategy constant to strategy name.
    SAMPLING_LOOKUP = {
        NONE: b"none",
        AUTO: b"auto",
        PERIOD: b"period",
        EVENT: b"event",
        DIFFERENTIAL: b"differential",
        EVENT_RATE: b"event-rate",
        DIFFERENTIAL_RATE: b"differential-rate",
    }

    # SAMPLING_LOOKUP not found by pylint
    #
    # pylint: disable-msg = E0602

    ## @brief Mapping from strategy name to strategy constant.
    SAMPLING_LOOKUP_REV = dict((v, k) for k, v in SAMPLING_LOOKUP.items())

    # pylint: enable-msg = E0602

    OBSERVE_UPDATES = False
    "True if a strategy must be attached to its sensor as an observer"

    def __init__(self, inform_callback, sensor, *params, **kwargs):
        self.ioloop = kwargs.get('ioloop') or tornado.ioloop.IOLoop.current()
        self._inform_callback = inform_callback
        self._sensor = sensor
        self._params = params

    @classmethod
    def get_strategy(cls, strategyName, inform_callback, sensor,
                     *params, **kwargs):
        """Factory method to create a strategy object.

        Parameters
        ----------
        strategyName : str or bytes
            Name of strategy.
        inform_callback : callable, signature inform_callback(sensor, reading)
            Callback to receive inform messages.
        sensor : Sensor object
            Sensor to sample.
        params : list of objects
            Custom sampling parameters for specified strategy.

        Keyword Arguments
        -----------------
        ioloop : tornado.ioloop.IOLoop instance, optional
            Tornado ioloop to use, otherwise tornado.ioloop.IOLoop.current()

        Returns
        -------
        strategy : :class:`SampleStrategy` object
            The created sampling strategy.

        """
        if is_text(strategyName):
            strategyName = strategyName.encode('ascii')
        if strategyName not in cls.SAMPLING_LOOKUP_REV:
            raise ValueError("Unknown sampling strategy '%s'. "
                             "Known strategies are %s."
                             % (strategyName, list(cls.SAMPLING_LOOKUP.values())))

        strategyType = cls.SAMPLING_LOOKUP_REV[strategyName]
        if strategyType == cls.NONE:
            return SampleNone(inform_callback, sensor, *params, **kwargs)
        elif strategyType == cls.AUTO:
            return SampleAuto(inform_callback, sensor, *params, **kwargs)
        elif strategyType == cls.EVENT:
            return SampleEvent(inform_callback, sensor, *params, **kwargs)
        elif strategyType == cls.DIFFERENTIAL:
            return SampleDifferential(inform_callback, sensor,
                                      *params, **kwargs)
        elif strategyType == cls.PERIOD:
            return SamplePeriod(inform_callback, sensor, *params, **kwargs)
        elif strategyType == cls.EVENT_RATE:
            return SampleEventRate(inform_callback, sensor, *params, **kwargs)
        elif strategyType == cls.DIFFERENTIAL_RATE:
            return SampleDifferentialRate(inform_callback, sensor,
                                          *params, **kwargs)

    def update(self, sensor, reading):
        """Callback used by the sensor's notify() method.

        This update method is called whenever the sensor value is set
        so sensor will contain the right info. Note that the strategy
        does not really need to be passed a sensor because it already
        has a handle to it, but receives it due to the generic observer
        mechanism.

        Sub-classes should override this method or :meth:`start` to provide
        the necessary sampling strategy. Sub-classes should also ensure that
        :meth:`update` is thread-safe; an easy way to do this is by using
        the @update_in_ioloop decorator.

        Parameters
        ----------
        sensor : Sensor object
            The sensor which was just updated.
        reading : (timestamp, status, value) tuple
            Sensor reading as would be returned by sensor.read()

        """
        pass

    def get_sampling(self):
        """The Strategy constant for this sampling strategy.

        Sub-classes should implement this method and return the
        appropriate constant.

        Returns
        -------
        strategy : Strategy constant
            The strategy type constant for this strategy.

        """
        raise NotImplementedError

    def get_sampling_formatted(self):
        """The current sampling strategy and parameters.

        The strategy is returned as a byte string and the values
        in the parameter list are formatted as byte strings using
        the formatter for this sensor type.

        Returns
        -------
        strategy_name : bytes
            KATCP name for the strategy.
        params : list of bytes
            KATCP formatted parameters for the strategy.

        """
        strategy = self.get_sampling()
        strategy = self.SAMPLING_LOOKUP[strategy]
        params = []
        for param in self._params:
            if not is_bytes(param):
                if PY2:
                    param = str(param)
                else:
                    param = str(param).encode('ascii')
            params.append(param)
        return strategy, params

    def attach(self):
        """Attach strategy to its sensor and send initial update."""
        s = self._sensor
        self.update(s, s.read())
        self._sensor.attach(self)

    def detach(self):
        """Detach strategy from its sensor."""
        self._sensor.detach(self)

    def cancel(self):
        """Detach strategy from its sensor and cancel ioloop callbacks."""
        if self.OBSERVE_UPDATES:
            self.detach()
        self.ioloop.add_callback(self.cancel_timeouts)

    def start(self):
        """Start operating the strategy.

        Subclasses that override start() should call the super method before
        it does anything that uses the ioloop. This will attach to the sensor
        as an observer if :attr:`OBSERVE_UPDATES` is True, and sets
        :attr:`_ioloop_thread_id` using `thread.get_ident()`.

        """
        def first_run():
            self._ioloop_thread_id = get_thread_ident()
            if self.OBSERVE_UPDATES:
                self.attach()
        self.ioloop.add_callback(first_run)

    def inform(self, reading):
        """Inform strategy creator of the sensor status."""
        try:
            self._inform_callback(self._sensor, reading)
        except Exception:
            log.exception('Unhandled exception trying to send {!r} '
                          'for sensor {!r} of type {!r}'
                          .format(reading, self._sensor.name, self._sensor.type))

    def cancel_timeouts(self):
        """Override this method to cancel any outstanding ioloop timeouts."""
        pass


class SampleAuto(SampleStrategy):
    """Strategy which sends updates whenever the sensor itself is updated."""

    OBSERVE_UPDATES = True

    def __init__(self, inform_callback, sensor, *params, **kwargs):
        SampleStrategy.__init__(self, inform_callback, sensor, *params, **kwargs)
        if params:
            raise ValueError("The 'auto' strategy takes no parameters.")

    @update_in_ioloop
    def update(self, sensor, reading):
        self.inform(reading)

    def get_sampling(self):
        return SampleStrategy.AUTO


class SampleNone(SampleStrategy):
    """Sampling strategy which never sends any updates."""

    def __init__(self, inform_callback, sensor, *params, **kwargs):
        SampleStrategy.__init__(self, inform_callback, sensor, *params, **kwargs)
        if params:
            raise ValueError("The 'none' strategy takes no parameters.")

    def start(self):
        # Do nothing, since we don't need set anything up
        pass

    def get_sampling(self):
        return SampleStrategy.NONE


class SampleDifferential(SampleStrategy):
    """Differential sampling strategy for integer and float sensors.

    Sends updates only when the value has changed by more than some
    specified threshold, or the status changes.

    """

    OBSERVE_UPDATES = True

    def __init__(self, inform_callback, sensor, *params, **kwargs):
        SampleStrategy.__init__(self, inform_callback, sensor, *params, **kwargs)
        if len(params) != 1:
            raise ValueError("The 'differential' strategy takes "
                             "one parameter.")
        if sensor._sensor_type not in (Sensor.INTEGER, Sensor.FLOAT,
                                       Sensor.TIMESTAMP):
            raise ValueError("The 'differential' strategy is only valid for "
                             "float, integer and timestamp sensors.")
        if sensor._sensor_type == Sensor.INTEGER:
            self._threshold = int(params[0])
            if self._threshold <= 0:
                raise ValueError("The diff amount must be a positive integer.")
        elif sensor._sensor_type == Sensor.FLOAT:
            self._threshold = float(params[0])
            if self._threshold <= 0:
                raise ValueError("The diff amount must be a positive float.")
        else:
            # _sensor_type must be Sensor.TIMESTAMP

            # There is a potential snafu here if katcpv4 server is used, since
            # the timestamp sensor type should be in milliseconds. For now, just
            # ignore this eventuality, and fix if anyone actually needs this
            self._threshold = float(params[0])
            if self._threshold <= 0:
                raise ValueError("The diff amount must be a positive number "
                                 "of seconds.")
        self._lastStatus = None
        self._lastValue = None

    @update_in_ioloop
    def update(self, sensor, reading):
        _timestamp, status, value = reading
        if (status != self._lastStatus or
                abs(value - self._lastValue) > self._threshold):
            self._lastStatus = status
            self._lastValue = value
            self.inform(reading)

    def get_sampling(self):
        return SampleStrategy.DIFFERENTIAL


class SamplePeriod(SampleStrategy):
    """Periodic sampling strategy."""

    def __init__(self, inform_callback, sensor, *params, **kwargs):
        SampleStrategy.__init__(self, inform_callback, sensor, *params, **kwargs)
        if len(params) != 1:
            raise ValueError("The 'period' strategy takes one parameter. "
                             "Parameters passed: %r, in pid : %s"
                             % (params, os.getpid()))
        period = float(params[0])
        if period <= 0:
            raise ValueError("The period must be a positive float in seconds. "
                             "Parameters passed: %r, in pid : %s"
                             % (params, os.getpid()))
        self._period = period

    def start(self):
        super(SamplePeriod, self).start()
        def start_periodic_sampling():
            self.next_time = self.ioloop.time()
            self._run_once()
        self.ioloop.add_callback(start_periodic_sampling)

    def _run_once(self):
        assert get_thread_ident() == self._ioloop_thread_id
        now = self.ioloop.time()
        self.inform(self._sensor.read())
        self.next_time += self._period
        if self.next_time < now:
            # Catch up if we have fallen far behind
            self.next_time = now + self._period
        self.next_timeout_handle = self.ioloop.call_at(self.next_time,
                                                       self._run_once)

    def get_sampling(self):
        return SampleStrategy.PERIOD

    def cancel_timeouts(self):
        self.ioloop.remove_timeout(self.next_timeout_handle)


class SampleEventRate(SampleStrategy):
    """Event rate sampling strategy.

    Report the sensor value whenever it changes or if more than
    *longest_period* seconds have passed since the last reported
    update. However, do not report the value if less than
    *shortest_period* seconds have passed since the last reported
    update.

    """

    OBSERVE_UPDATES = True

    def __init__(self, inform_callback, sensor, *params, **kwargs):
        SampleStrategy.__init__(self, inform_callback, sensor, *params, **kwargs)
        if len(params) != 2:
            raise ValueError("The 'event-rate' strategy takes two parameters.")
        shortest_period = float(params[0])
        longest_period = float(params[1])
        self._next_periodic = 1e99

        if not 0 <= shortest_period <= longest_period:
            raise ValueError("The shortest and longest periods must "
                             "satisfy 0 <= shortest_period <= longest_period")
        self._shortest_period = shortest_period
        self._short_timeout_handle = None
        self._longest_period = longest_period
        # don't send updates until timestamp _not_before
        self._not_before = 0
        self._last_reading_sent = (None, None, None)

    def inform(self, reading):
        SampleStrategy.inform(self, reading)
        now = self.ioloop.time()
        self._not_before = now + self._shortest_period
        self._not_after = now + self._longest_period
        self._last_reading_sent = reading

    def start(self):
        super(SampleEventRate, self).start()
        self.ioloop.add_callback(self._periodic_sampling)

    def _periodic_sampling(self):
        if self._not_after >= AGE_OF_UNIVERSE:
            # Ignore stupidly long periods
            self._periodic_timeout_handle = None
            return
        if self.ioloop.time() >= self._not_after:
            self.inform(self._sensor.read())
        # We depend on self.inform() having updated self._not_after
        self._periodic_timeout_handle = self.ioloop.call_at(
            self._not_after, self._periodic_sampling)

    def _short_timeout_handler(self):
        self._short_timeout_handle = None
        if (self.ioloop.time() >= self._not_before):
            self.inform(self._sensor.read())

    def _sensor_changed(self, reading):
        _, status, value = reading
        _, last_s, last_v = self._last_reading_sent
        return status != last_s or value != last_v

    @update_in_ioloop
    def update(self, sensor, reading):
        sensor_changed = self._sensor_changed(reading)
        if not sensor_changed:
            # Ignore update if sensor value/status is unchanged
            return
        now = self.ioloop.time()
        if now < self._not_before:
            # Too soon to send an update again
            if not self._short_timeout_handle:
                # Make sure we schedule a callback to send the sensor value as
                # soon as the minimum period since the last update has expired
                self._short_timeout_handle = self.ioloop.call_at(
                    self._not_before, self._short_timeout_handler)
        else:
            # Send the sensor value, updating self._not_before and
            # self._last_reading_sent in the process
            self.inform(reading)

    def get_sampling(self):
        return SampleStrategy.EVENT_RATE

    def cancel_timeouts(self):
        if self._periodic_timeout_handle:
            self.ioloop.remove_timeout(self._periodic_timeout_handle)
        if self._short_timeout_handle:
            self.ioloop.remove_timeout(self._short_timeout_handle)


class SampleEvent(SampleEventRate):
    """Strategy which sends updates when the sensor value or status changes.

    Since SampleEvent is just a special case of SampleEventRate, we use
    SampleEventRate with the appropriate default values to implement
    SampleEvent.

    """
    def __init__(self, inform_callback, sensor, *params, **kwargs):
        if len(params) > 0:
            raise ValueError("The 'event' strategy takes no parameters.")
        super(SampleEvent, self).__init__(inform_callback, sensor, 0, 1e99, **kwargs)
        # Fix up the parameters so we don't see the extra parameters that were
        # passed to SampleEventRate
        self._params = params

    def get_sampling(self):
        return SampleStrategy.EVENT


class SampleDifferentialRate(SampleEventRate):
    """Differential rate sampling strategy.

    Report the value whenever it changes by more than *difference*
    from the last reported value or if more than *longest_period*
    seconds have passed since the last reported update. However, do
    not report the value until *shortest_period* seconds have passed
    since the last reported update. The behaviour if *shortest_period*
    is greater than *longest_period* is undefined. May only be
    implemented for float and integer sensors.

    """
    def __init__(self, inform_callback, sensor, *params, **kwargs):
        # Remove the 'difference' parameter from params for the super call
        if len(params) != 3:
            raise ValueError("The 'differential-rate' strategy takes three parameters.")
        super(SampleDifferentialRate, self).__init__(inform_callback, sensor,
                                                     *params[1:], **kwargs)
        # Fix up the parameters so we do not lose the params that are not
        # passed to SampleEventRate
        self._params = params

        difference = params[0]
        if sensor.stype == 'integer':
            difference = int(difference)
        elif sensor.stype == 'float':
            difference = float(difference)
        else:
            raise ValueError('The differential-rate strategy can only be '
                             'defined for integer or float sensors')
        self.difference = difference
        # Initial value that should not cause errors in _sensor_changed(),
        # but should let it return True
        self._last_reading_sent = (None, None, 1e99)

    def get_sampling(self):
        return SampleStrategy.DIFFERENTIAL_RATE

    def _sensor_changed(self, reading):
        _, status, value = reading
        _, last_s, last_v = self._last_reading_sent
        return (abs(value - last_v) > self.difference or status != last_s)



# Copyright 2015 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

from builtins import object

import tornado.concurrent

from _thread import get_ident as get_thread_ident
from tornado.concurrent import Future
from tornado.gen import Return

from katcp import Sensor, client, kattypes, resource, server
from katcp.core import AttrDict, Message, ProtocolFlags, convert_method_name


def fake_KATCP_client_resource_factory(
        KATCPClientResourceClass, fake_options, resource_spec, *args, **kwargs):
    """Create a fake KATCPClientResource-like class and a fake-manager

    Parameters
    ----------
    KATCPClientResourceClass : class
        Subclass of :class:`katcp.resource_client.KATCPClientResource`
    fake_options : dict
        Options for the faking process. Keys:
            allow_any_request : bool, default False
            (TODO not implemented behaves as if it were True)
    resource_spec, *args, **kwargs : passed to KATCPClientResourceClass

    A subclass of the passed-in KATCPClientResourceClass is created that replaces the
    internal InspecingClient instances with fakes using fake_inspecting_client_factory()
    based on the InspectingClient class used by KATCPClientResourceClass.

    Returns
    -------
    (fake_katcp_client_resource, fake_katcp_client_resource_manager):

    fake_katcp_client_resource : instance of faked subclass of KATCPClientResourceClass
    fake_katcp_client_resource_manager : :class:`FakeKATCPClientResourceManager` instance
        Bound to the `fake_katcp_client_resource` instance.

    """
    # TODO Implement allow_any_request functionality. When True, any unknown request (even
    # if there is no fake implementation) should succeed
    allow_any_request = fake_options.get('allow_any_request', False)

    class FakeKATCPClientResource(KATCPClientResourceClass):
        def inspecting_client_factory(self, host, port, ioloop_set_to):
            real_instance = (super(FakeKATCPClientResource, self)
                             .inspecting_client_factory(host, port, ioloop_set_to) )
            fic, fic_manager = fake_inspecting_client_factory(
                real_instance.__class__, fake_options, host, port,
                ioloop=ioloop_set_to, auto_reconnect=self.auto_reconnect)
            self.fake_inspecting_client_manager = fic_manager
            return fic

    fkcr = FakeKATCPClientResource(resource_spec, *args, **kwargs)
    fkcr_manager = FakeKATCPClientResourceManager(fkcr)
    return (fkcr, fkcr_manager)

def fake_KATCP_client_resource_container_factory(
        KATCPClientResourceContainerClass, fake_options, resources_spec,
        *args, **kwargs):
    """Create a fake KATCPClientResourceContainer-like class and a fake-manager

    Parameters
    ----------
    KATCPClientResourceContainerClass : class
        Subclass of :class:`katcp.resource_client.KATCPClientResourceContainer`
    fake_options : dict
        Options for the faking process. Keys:
            allow_any_request : bool, default False
            (TODO not implemented behaves as if it were True)
    resources_spec, *args, **kwargs : passed to KATCPClientResourceContainerClass

    A subclass of the passed-in KATCPClientResourceClassContainer is created that replaces the
    KATCPClientResource child instances with fakes using fake_KATCP_client_resource_factory()
    based on the KATCPClientResource class used by `KATCPClientResourceContainerClass`.

    Returns
    -------
    (fake_katcp_client_resource_container, fake_katcp_client_resource_container_manager):

    fake_katcp_client_resource_container : instance of faked subclass of
                                           KATCPClientResourceContainerClass
    fake_katcp_client_resource_manager : :class:`FakeKATCPClientResourceContainerManager`
                                         instance
        Bound to the `fake_katcp_client_resource_container` instance.

    """
    # TODO allow_any_request see comment in fake_KATCP_client_resource_factory()
    allow_any_request = fake_options.get('allow_any_request', False)

    class FakeKATCPClientResourceContainer(KATCPClientResourceContainerClass):
        def __init__(self, *args, **kwargs):
            self.fake_client_resource_managers = {}
            super(FakeKATCPClientResourceContainer, self).__init__(*args, **kwargs)

        def client_resource_factory(self, res_spec, parent, logger):
            real_instance = (super(FakeKATCPClientResourceContainer, self)
                             .client_resource_factory(res_spec, parent, logger) )
            fkcr, fkcr_manager = fake_KATCP_client_resource_factory(
                real_instance.__class__, fake_options,
                res_spec, parent=self, logger=logger)
            self.fake_client_resource_managers[
                resource.escape_name(fkcr.name)] = fkcr_manager
            return fkcr

    fkcrc = FakeKATCPClientResourceContainer(resources_spec, *args, **kwargs)
    fkcrc_manager = FakeKATCPClientResourceContainerManager(fkcrc)
    return (fkcrc, fkcrc_manager)


class FakeKATCPClientResourceManager(object):
    """Manage a fake KATCPClientResource instance"""
    def __init__(self, fake_katcp_resource_client):
        self._fkrc = fake_katcp_resource_client

    @property
    def _fic_manager(self):
        try:
           fic_manager = self._fkrc.fake_inspecting_client_manager
        except AttributeError:
            raise RuntimeError(
                'Fake KATCPClientResource must be started before it can be managed')
        return fic_manager

    @property
    def fake_sensor_infos(self):
        return self._fic_manager.fake_sensor_infos

    def add_sensors(self, sensor_infos):
        """Add fake sensors

        sensor_infos is a dict <sensor-name> : (
            <description>, <unit>, <sensor-type>, <params>*)
        The sensor info is string-reprs of whatever they are, as they would be on
        the wire in a real KATCP connection. Values are passed to a katcp.Message object,
        so some automatic conversions are done, hence it is OK to pass numbers without
        stringifying them.

        """
        self._fic_manager.add_sensors(sensor_infos)

    def add_request_handlers_object(self, rh_obj):
        """Add fake request handlers from an object with request_* method(s)

        See :class:`FakeInspectingClientManager` for detail
        """
        return self._fic_manager.add_request_handlers_object(rh_obj)

    def add_request_handlers_dict(self, rh_dict):
        """Add fake request handlers from a dict keyed by request name

        See :class:`FakeInspectingClientManager` for detail
        """
        return self._fic_manager.add_request_handlers_dict(rh_dict)


class FakeKATCPClientResourceContainerManager(object):
    def __init__(self, fake_katcp_resource_container):
        self._fkcrc = fake_katcp_resource_container

    @property
    def child_managers(self):
        return self._fkcrc.fake_client_resource_managers

    def add_sensors(self, child_name, sensor_infos):
        """Add fake sensors

        child_name : str
            Name of the client to which the sensor should be added as used in the keys of
            the FakeKATCPClientResourceContainer `children` container.
        sensor_infos is a dict <sensor-name> : (
            <description>, <unit>, <sensor-type>, <params>*)
        The sensor info is string-reprs of whatever they are, as they would be on
        the wire in a real KATCP connection. Values are passed to a katcp.Message object,
        so some automatic conversions are done, hence it is OK to pass numbers without
        stringifying them.

        """
        self.child_managers[child_name].add_sensors(sensor_infos)

    def add_request_handlers_object(self, child_name, rh_obj):
        """Add fake request handlers from an object with request_* method(s)

        See :class:`FakeInspectingClientManager` for detail
        """
        return self.child_managers[child_name].add_request_handlers_object(rh_obj)

    def add_request_handlers_dict(self, child_name, rh_dict):
        """Add fake request handlers from a dict keyed by request name

        See :class:`FakeInspectingClientManager` for detail
        """
        return self.child_managers[child_name].add_request_handlers_dict(rh_dict)

def fake_inspecting_client_factory(InspectingClass, fake_options, host, port,
                                   *args, **kwargs):
    # Note, that this mechanism makes many assumptions about the internals of
    # InspectingClientAsync and DeviceClient, so it may well break if substantial changes
    # are made to the implementation of either.

    allow_any_request = fake_options.get('allow_any_request', False)
    # TODO Implement any request stuff. See comment above in
    # fake_KATCP_client_resource_factory


    class FakeInspectingClass(InspectingClass):
        def inform_hook_client_factory(self, host, port, *args, **kwargs):
            fac = FakeAsyncClient(host, port, *args, **kwargs)
            # We'll be ignoring the inform hooks, rather stimulating the InspectingClass at a
            # higher level
            fac.hook_inform = lambda *x, **xx : None
            return fac

    fic = FakeInspectingClass(host, port, *args, **kwargs)
    fic_manager = FakeInspectingClientManager(fic)
    return (fic, fic_manager)


class FakeInspectingClientManager(object):

    request_help = server.DeviceServer.__dict__["request_help"]

    def __init__(self, fake_inspecting_client):
        self._fic = fake_inspecting_client
        self._fkc = fake_inspecting_client.katcp_client
        self.add_request_handlers_object(self)
        self.fake_sensor_infos = {}

    @kattypes.return_reply(kattypes.Int())
    def request_sensor_list(self, req, msg):
        """Sensor list"""
        if msg.arguments:
            name = (msg.arguments[0],)
            keys = (name, )
            if name not in self.fake_sensor_infos:
                return ("fail", "Unknown sensor name.")
        else:
            keys = self.fake_sensor_infos.keys()

        num_informs = 0
        for sensor_name in keys:
            infos = self.fake_sensor_infos[sensor_name]
            num_informs += 1
            req.inform(sensor_name, *infos)

        return ('ok', num_informs)

    @property
    def _request_handlers(self):
        """For compatibility with methods stolen from server.DeviceServer"""
        return self._fkc.request_handlers

    def add_sensors(self, sensor_infos):
        """Add fake sensors

        sensor_infos is a dict <sensor-name> : (
            <description>, <unit>, <sensor-type>, <params>*)
        The sensor info is string-reprs of whatever they are, as they would be on
        the wire in a real KATCP connection. Values are passed to a katcp.Message object,
        so some automatic conversions are done, hence it is OK to pass numbers without
        stringifying them.

        """
        # Check sensor validity. parse_type() and parse_params should raise if not OK
        for s_name, s_info in sensor_infos.items():
            s_type = Sensor.parse_type(s_info[2])
            s_params = Sensor.parse_params(s_type, s_info[3:])
        self.fake_sensor_infos.update(sensor_infos)
        self._fic._interface_changed.set()

    def add_request_handlers_object(self, rh_obj):
        """Add fake request handlers from an object with request_* method(s)

        See :meth:`FakeInspectingClientManager.add_request_handlers_dict` for more detail.

        """
        rh_dict = {}
        for name in dir(rh_obj):
            if not callable(getattr(rh_obj, name)):
                continue
            if name.startswith("request_"):
                request_name = convert_method_name("request_", name)
                req_meth = getattr(rh_obj, name)
                rh_dict[request_name] = req_meth

        self.add_request_handlers_dict(rh_dict)

    def add_request_handlers_dict(self, rh_dict):
        """Add fake request handler functions from a dict keyed by request name

        Note the keys must be the KATCP message name (i.e. "the-request", not
        "the_request")

        The request-handler interface is more or less compatible with request handler API
        in :class:`katcp.server.DeviceServerBase`. The fake ClientRequestConnection req
        object does not support mass_inform() or reply_with_message.

        """
        # Check that all the callables have docstrings as strings
        for req_func in rh_dict.values():
            assert req_func.__doc__, "Even fake request handlers must have docstrings"
        self._fkc.request_handlers.update(rh_dict)
        self._fic._interface_changed.set()


class FakeKATCPServerError(Exception):
    """Raised if a FakeKATCPServer is used in an unsupported way"""


class FakeKATCPServer(object):
    """Fake the parts of a KATCP server used by katcp.server.ClientConnection"""
    def get_address(self, conn_id):
        return '<fake-client-connection: {!r}>'.format(conn_id)

    def send_message(self, conn_id, msg):
        raise FakeKATCPServerError(
            'Cannot send messages via fake request/connection object')

    def mass_send_message(self, msg):
        raise FakeKATCPServerError(
            'Cannot send messages via fake request/connection object')

    def flush_on_close(self):
        f = Future()
        f.set_result(None)
        return f


class FakeClientRequestConnection(server.ClientRequestConnection):
    def __init__(self, *args, **kwargs):
        super(FakeClientRequestConnection, self).__init__(*args, **kwargs)
        self.informs_sent = []

    def inform(self, *args):
        inf_msg = Message.reply_inform(self.msg, *args)
        self.informs_sent.append(inf_msg)


class FakeAsyncClient(client.AsyncClient):
    """Fake version of :class:`katcp.client.AsyncClient`

    Useful for testing and simulation

    By default assumes that the client is connected to a fully featured KATCPv5
    server. Call preset_protocol_flags() to override
    """

    def __init__(self, *args, **kwargs):
        self.request_handlers = {}
        self.fake_server = FakeKATCPServer()
        self.client_connection = server.ClientConnection(
            self.fake_server, 'fake-async-client')
        super(FakeAsyncClient, self).__init__(*args, **kwargs)

    @tornado.gen.coroutine
    def _install(self):
        self.ioloop_thread_id = get_thread_ident()
        self._running.set()
        yield self._connect()

    @tornado.gen.coroutine
    def _connect(self):
        assert get_thread_ident() == self.ioloop_thread_id
        self._stream = AttrDict(error=None)
        self._connected.set()
        self.notify_connected(True)
        self.preset_protocol_flags(ProtocolFlags(5, 0, set('IM')))

    def _disconnect(self):
        if self._stream:
            self._stream_closed_callback(self._stream)

    @tornado.gen.coroutine
    def future_request(self, msg, timeout=None, use_mid=None):
        """Send a request message, with future replies.

        Parameters
        ----------
        msg : Message object
            The request Message to send.
        timeout : float in seconds
            How long to wait for a reply. The default is the
            the timeout set when creating the AsyncClient.
        use_mid : boolean, optional
            Whether to use message IDs. Default is to use message IDs
            if the server supports them.

        Returns
        -------
        A tornado.concurrent.Future that resolves with:

        reply : Message object
            The reply message received.
        informs : list of Message objects
            A list of the inform messages received.

        """
        mid = self._get_mid_and_update_msg(msg, use_mid)

        if msg.name in self.request_handlers:
            req = FakeClientRequestConnection(self.client_connection, msg)
            reply_msg = yield tornado.gen.maybe_future(
                self.request_handlers[msg.name](req, msg))
            reply_informs = req.informs_sent
        else:
            reply_msg = Message.reply(msg.name, 'ok')
            reply_informs = []

        reply_msg.mid = mid
        raise Return((reply_msg, reply_informs))



# test_client.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for client module."""

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import gc
import logging
import threading
import time
import unittest
import weakref

from builtins import range
from concurrent.futures import Future

import mock
import tornado
import tornado.testing

from tornado import gen

import katcp

from katcp.core import Message, ProtocolFlags
from katcp.testutils import (DeviceTestServer, TestLogHandler, TestUtilMixin,
                             TimewarpAsyncTestCase, WaitingMock,
                             counting_callback, start_thread_with_cleanup)

log_handler = TestLogHandler()
logging.getLogger("katcp").addHandler(log_handler)

logger = logging.getLogger(__name__)

# Number of requests on DeviceTestServer
NUM_HELP_MESSAGES = len(DeviceTestServer._request_handlers)


def remove_version_connect(msgs):
    """Remove #version-connect messages from a list of messages"""
    return [msg for msg in msgs if msg.name != 'version-connect']


class TestDeviceClientServerDetection(unittest.TestCase, TestUtilMixin):
    def setUp(self):
        self.client = katcp.DeviceClient('localhost', 0)
        self.assertFalse(self.client._received_protocol_info.isSet())
        self.v4_build_state = Message.inform('build-state', 'blah-5.21a3')
        self.v4_version = Message.inform('version', '7.3')
        self.v5_version_connect_mid = Message.inform(
            'version-connect', 'katcp-protocol', '5.0-I')
        self.v5_version_connect_nomid = Message.inform(
            'version-connect', 'katcp-protocol', '5.0')

    def _check_v5_mid(self):
        self.assertTrue(self.client._received_protocol_info.isSet())
        pf = self.client.protocol_flags
        self.assertTrue(pf.supports(pf.MESSAGE_IDS))
        self.assertTrue(self.client._server_supports_ids)
        self.assertEqual(pf.major, 5)
        self.assertEqual(self.client.convert_seconds(1), 1)

    def test_valid_v5(self):
        self.client.handle_message(self.v5_version_connect_mid)
        self._check_v5_mid()

    def _check_v4(self):
        self.assertTrue(self.client._received_protocol_info.isSet())
        pf = self.client.protocol_flags
        self.assertFalse(pf.supports(pf.MESSAGE_IDS))
        self.assertFalse(self.client._server_supports_ids)
        self.assertEqual(pf.major, 4)
        self.assertEqual(self.client.convert_seconds(1), 1000)

    def test_valid_v4_version_first(self):
        self.client.handle_message(self.v4_version)
        self._check_v4()
        self.client.handle_message(self.v4_build_state)
        self._check_v4()

    def test_valid_v4_build_state_first(self):
        self.client.handle_message(self.v4_build_state)
        self._check_v4()
        self.client.handle_message(self.v4_version)
        self._check_v4()

    def test_inconsistent_v4_then_v5(self):
        self.client.handle_message(self.v4_build_state)
        self._check_v4()
        self.client._disconnect = mock.Mock()
        self.client._logger = mock.Mock()
        self.client.handle_message(self.v5_version_connect_mid)
        self.assertEqual(self.client._disconnect.call_count, 1)
        self.assertEqual(self.client._auto_reconnect, False)
        self.assertEqual(self.client._logger.error.call_count, 1)
        log_msg = self.client._logger.error.call_args[0][0]
        self.assertIn('Protocol Version Error', log_msg)

    def test_inconsistent_v5_then_v4(self):
        self.client.handle_message(self.v5_version_connect_mid)
        self._check_v5_mid()
        self.client._disconnect = mock.Mock()
        self.client._logger = mock.Mock()
        self.client.handle_message(self.v4_build_state)
        self.assertEqual(self.client._disconnect.call_count, 1)
        self.assertEqual(self.client._auto_reconnect, False)
        self.assertEqual(self.client._logger.error.call_count, 1)
        log_msg = self.client._logger.error.call_args[0][0]
        self.assertIn('Protocol Version', log_msg)

    def test_preset(self):
        self.client.preset_protocol_flags(ProtocolFlags(
            5, 0, [ProtocolFlags.MESSAGE_IDS]))
        self._check_v5_mid()
        self.client._received_protocol_info.clear()
        self.client.preset_protocol_flags(ProtocolFlags(4, 0, ''))
        self._check_v4()

    def test_preset_v4_then_v5(self):
        self.client.preset_protocol_flags(ProtocolFlags(4, 0, ''))
        self._check_v4()
        self.client._disconnect = mock.Mock()
        self.client._logger = mock.Mock()
        # A version 5 version-connect message should result in a warning
        self.client.handle_message(self.v5_version_connect_mid)
        self.assertEqual(self.client._disconnect.call_count, 0)
        self.assertEqual(self.client._auto_reconnect, True)
        self.assertEqual(self.client._logger.warn.call_count, 1)
        log_msg = self.client._logger.warn.call_args[0][0]
        self.assertIn('Protocol Version', log_msg)
        self._check_v4()
        self.client._logger.warn.reset_mock()
        # Version 4-identifying informs should not cause any warnings
        self.client.handle_message(self.v4_version)
        self.client.handle_message(self.v4_build_state)
        self.assertEqual(self.client._disconnect.call_count, 0)
        self.assertEqual(self.client._auto_reconnect, True)
        self.assertEqual(self.client._logger.warn.call_count, 0)
        self._check_v4()

    def test_inform_version_connect(self):
        # Test that the inform handler doesn't screw up with a non-katcp related
        # version-connect inform.
        self.client.handle_message(Message.inform(
            'version-connect', 'not-katcp', '5.71a3'))
        # Should not raise any errors, but should also not set the protocol
        # infor received flag.
        self.assertFalse(self.client._received_protocol_info.isSet())

    def test_preset_v5_then_v4(self):
        def check_warn_not_disconect():
            self.assertEqual(self.client._disconnect.call_count, 0)
            self.assertEqual(self.client._auto_reconnect, True)
            self.assertEqual(self.client._logger.warn.call_count, 1)
            log_msg = self.client._logger.warn.call_args[0][0]
            self.assertIn('Protocol Version', log_msg)
            self.client._logger.warn.reset_mock()

        self.client.preset_protocol_flags(ProtocolFlags(
            5, 0, [ProtocolFlags.MESSAGE_IDS]))
        self._check_v5_mid()
        self.client._disconnect = mock.Mock()
        self.client._logger = mock.Mock()
        # Any Version 4-identifying informs should cause warnings
        self.client.handle_message(self.v4_build_state)
        check_warn_not_disconect()
        self._check_v5_mid()
        self.client.handle_message(self.v4_version)
        check_warn_not_disconect()
        self._check_v5_mid()
        # A version 5 version-connect message with different flags should also
        # result in a warning
        self.client.handle_message(self.v5_version_connect_nomid)
        check_warn_not_disconect()
        self._check_v5_mid()
        # An identical version 5 version-connect message should not result in
        # any warnings
        self.client.handle_message(self.v5_version_connect_mid)
        self.assertEqual(self.client._disconnect.call_count, 0)
        self.assertEqual(self.client._logger.warn.call_count, 0)
        self.assertEqual(self.client._auto_reconnect, True)
        self._check_v5_mid()


class TestDeviceClientIntegrated(unittest.TestCase, TestUtilMixin):

    def setUp(self):
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)

        host, port = self.server.bind_address

        self.client = katcp.DeviceClient(host, port)
        self.client.enable_thread_safety()
        start_thread_with_cleanup(self, self.client, start_timeout=1)
        self.client.wait_connected(timeout=1)

    def test_versions(self):
        """Test that the versions parameter is populated."""
        preamble_done = self.client.handle_reply = WaitingMock()
        # Do a request and wait for it to be done so that we can be sure we received the
        # full connection-header before testing
        self.client.request(Message.request('watchdog'))
        preamble_done.assert_wait_call_count(1, timeout=1)
        versions = self.client.versions
        self.assertIn('katcp', ' '.join(versions['katcp-library']))
        self.assertIn('device', ' '.join(versions['katcp-device']))
        self.assertTrue(' '.join(versions['katcp-protocol']))

    def test_request(self):
        """Test request method."""
        self.assertTrue(self.client.wait_protocol(1))
        self.client.send_request(Message.request("watchdog"))
        self.client._server_supports_ids = False
        with self.assertRaises(katcp.core.KatcpVersionError):
            self.client.send_request(Message.request("watchdog", mid=56))
        self.client._server_supports_ids = True
        self.client.send_request(Message.request("watchdog", mid=55))

        msgs = self.server.until_messages(2).result(timeout=1)
        self._assert_msgs_equal(msgs, [
            r"?watchdog",
            r"?watchdog[55]",
        ])

    def test_send_message(self):
        """Test send_message method."""
        self.client.send_message(Message.inform("random-inform"))

        msgs = self.server.until_messages(1).result(timeout=1)
        self._assert_msgs_equal(msgs, [
            r"#random-inform",
        ])

    def test_stop_and_restart(self):
        """Test stopping and then restarting a client."""
        self.client.wait_running(timeout=1)
        before_threads = threading.enumerate()
        self.client.stop(timeout=1)
        self.client.join(timeout=1)
        # Test that we have fewer threads than before
        self.assertLess(len(threading.enumerate()), len(before_threads))
        self.assertFalse(self.client._running.isSet())
        self.client.start(timeout=1)
        self.client.wait_running(timeout=1)
        # Now we should have the original number of threads again
        self.assertEqual(len(threading.enumerate()), len(before_threads))

    def test_is_connected(self):
        """Test is_connected method."""
        self.assertTrue(self.client.is_connected())
        # Use client.notify_connected to synchronise to the disconnection
        disconnected = threading.Event()
        self.client.notify_connected = (
            lambda connected: disconnected.set() if not connected else None)
        self.server.stop(timeout=1.)
        # Restart server during cleanup to keep teardown happy
        self.addCleanup(self.server.start)
        self.server.join(timeout=1.)
        # Wait for the client to be disconnected
        disconnected.wait(1.5)
        self.assertFalse(self.client.is_connected())

    def test_wait_connected(self):
        """Test wait_connected method."""
        start = time.time()
        # Ensure that we are connected at the start of the test
        self.assertTrue(self.client.wait_connected(1.0))
        # Check that we did not wait more than the timeout.
        self.assertTrue(time.time() - start < 1.0)
        # Now we will cause the client to disconnect by stopping the server, and then
        # checking that wait_connected returns fails, and waits approximately the right
        # amount of time
        # Use client.notify_connected to synchronise to the disconnection
        disconnected = threading.Event()
        self.client.notify_connected = (
            lambda connected: disconnected.set() if not connected else None)
        self.server.stop(timeout=0.1)
        self.server.join(timeout=1)
        # Restart server during cleanup to keep teardown happy
        self.addCleanup(self.server.start)
        # Wait for the client to be disconnected
        disconnected.wait(1)
        # Now check that wait_connected returns false
        start = time.time()
        self.assertFalse(self.client.wait_connected(0.1))
        # And waited more or less the timeout.
        self.assertTrue(0.05 < time.time() - start <= 0.15)

    def test_bad_socket(self):
        """Test what happens when select is called on a dead socket."""
        # wait for client to connect
        self.client.wait_connected(timeout=1)
        f = Future()
        def notify_connected(connected):
            if connected:
                f.set_result(connected)
        self.client.notify_connected = notify_connected

        # close stream while the client isn't looking
        # then wait for the client to notice
        stream = self.client._stream
        sockname = stream.socket.getpeername()
        self.client.ioloop.add_callback(stream.close)
        f.result(timeout=1.25)

        # check that client reconnected
        self.assertTrue(stream is not self.client._stream,
                        "Expected %r to not be %r" % (stream, self.client._stream))
        self.assertEqual(sockname, self.client._stream.socket.getpeername())

    def test_async_events_on_correct_ioloop(self):
        client = self.client
        # for managed ioloop, client must be using a different ioloop
        self.assertNotEqual(client.ioloop, tornado.ioloop.IOLoop.current())
        # check AsyncEvents (test is a little brittle)
        self.assertEqual(client.ioloop, client._stopped._ioloop)
        self.assertEqual(client.ioloop, client._running._ioloop)
        self.assertEqual(client.ioloop, client._waiting_to_retry._ioloop)
        self.assertEqual(client.ioloop, client._connected._ioloop)
        self.assertEqual(client.ioloop, client._disconnected._ioloop)
        self.assertEqual(client.ioloop, client._received_protocol_info._ioloop)


class TestDeviceClientMemoryLeaks(tornado.testing.AsyncTestCase):

    def setUp(self):
        super(TestDeviceClientMemoryLeaks, self).setUp()
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server, start_timeout=0.1)
        self.host, self.port = self.server.bind_address

    def use_on_managed_ioloop(self, client, timeout=0.1):
        client.start(timeout=timeout)
        client.wait_protocol(timeout=timeout)
        self.assertTrue(client.protocol_flags)
        client.stop(timeout=timeout)
        client.join(timeout=timeout)

    @gen.coroutine
    def use_on_unmanaged_ioloop(self, client, timeout=0.1):
        client.set_ioloop(self.io_loop)
        client.start()  # no timeout if starting from ioloop thread
        yield client.until_protocol(timeout=timeout)
        self.assertTrue(client.protocol_flags)
        client.stop(timeout=timeout)
        yield client.until_stopped(timeout=timeout)

    def test_no_memory_leak_managed_ioloop(self):
        client = katcp.DeviceClient(self.host, self.port)
        wr = weakref.ref(client)

        self.use_on_managed_ioloop(client)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    @tornado.testing.gen_test
    def test_no_memory_leak_unmanaged_ioloop(self):
        client = katcp.DeviceClient(self.host, self.port)
        wr = weakref.ref(client)

        # use test's ioloop, so client does not create its own (i.e., unmanaged)
        yield self.use_on_unmanaged_ioloop(client)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    @tornado.testing.gen_test
    def test_no_memory_leak_change_ioloop(self):
        client = katcp.DeviceClient(self.host, self.port)
        wr = weakref.ref(client)

        # start and stop client with managed ioloop
        self.use_on_managed_ioloop(client)

        # repeat with managed ioloop (new ioloop instance created)
        self.use_on_managed_ioloop(client)

        # change to unmanaged ioloop
        yield self.use_on_unmanaged_ioloop(client)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    def test_no_memory_leak_stop_current_ioloop(self):
        # The DeviceClient may use the current ioloop during initialisation.
        # Similarly, the DeviceTestServer created in setUp() may have references to
        # the current ioloop.  We make a new instance to replace the current
        # ioloop, which we can stop without affecting the server.  This test aims
        # to verify that the client is not dependent on the ioloop that was
        # current at the time the client was initialised.  This is only for
        # the "managed" ioloop case.
        ioloop = tornado.ioloop.IOLoop()
        tornado.ioloop.IOLoop.make_current(ioloop)

        client = katcp.DeviceClient(self.host, self.port)
        wr = weakref.ref(client)
        # close ioloop that was current when client created
        ioloop.close()
        self.use_on_managed_ioloop(client)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    def test_no_memory_leak_async_client(self):
        client = katcp.AsyncClient(self.host, self.port)
        wr = weakref.ref(client)

        self.use_on_managed_ioloop(client)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    def test_no_memory_leak_callback_client(self):
        client = katcp.CallbackClient(self.host, self.port)
        wr = weakref.ref(client)

        self.use_on_managed_ioloop(client)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    def test_no_memory_leak_stopped_server(self):
        client = katcp.DeviceClient(self.host, self.port)
        wr = weakref.ref(client)

        client.auto_reconnect_delay = 0.01
        client.start(timeout=0.1)
        client.wait_protocol(timeout=0.1)
        self.assertTrue(client.protocol_flags)
        # stop server before client
        self.server.stop()
        client.wait_disconnected(timeout=0.1)
        self.assertFalse(client.is_connected())
        client.stop(timeout=0.1)
        client.join(timeout=0.1)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    def test_no_memory_leak_invalid_server_no_reconnect(self):
        bad_port = 0
        client = katcp.DeviceClient(self.host, bad_port, auto_reconnect=False)
        wr = weakref.ref(client)

        client.start(timeout=0.1)
        client.wait_connected(timeout=0.1)
        self.assertFalse(client.is_connected())
        client.stop(timeout=0.1)
        client.join(timeout=0.1)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())

    def test_no_memory_leak_invalid_server_auto_reconnect(self):
        bad_port = 0
        client = katcp.DeviceClient(self.host, bad_port, auto_reconnect=True)
        wr = weakref.ref(client)

        client.auto_reconnect_delay = 0.01  # speed up test
        client.start(timeout=0.1)
        client.wait_connected(timeout=2 * client.auto_reconnect_delay)
        self.assertFalse(client.is_connected())
        client.stop(timeout=0.1)
        client.join(timeout=0.1)

        # clear strong reference and check if object can be garbage collected
        client = None
        gc.collect()
        self.assertIsNone(wr())


class TestBlockingClient(unittest.TestCase):
    def setUp(self):
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)

        host, port = self.server.bind_address

        self.client = katcp.BlockingClient(host, port)
        start_thread_with_cleanup(self, self.client, start_timeout=1)
        self.assertTrue(self.client.wait_protocol(timeout=1))

    def test_blocking_request(self):
        """Test blocking_request."""
        reply, informs = self.client.blocking_request(
            Message.request("watchdog"))
        self.assertEqual(reply.name, "watchdog")
        self.assertEqual(reply.arguments, [b"ok"])
        self.assertEqual(remove_version_connect(informs), [])

        reply, informs = self.client.blocking_request(
            Message.request("help"))
        self.assertEqual(reply.name, "help")
        self.assertEqual(reply.arguments, [b"ok", b"%d" % NUM_HELP_MESSAGES])
        self.assertEqual(len(informs), int(reply.arguments[1]))

    def test_blocking_request_mid(self):
        ## Test that the blocking client does the right thing with message
        ## identifiers
        # Wait for the client to detect the server protocol. Server should
        # support message identifiers
        self.assertTrue(self.client.wait_protocol(0.2))
        # Replace send_message so that we can check the message
        self.client.send_message = mock.Mock()

        # blocking_request() raises RuntimeError on timeout, so make a wrapper
        # function to eat that

        # TODO (NM 2014-09-08): Can remove this once we've made all clients handle
        # blocking_request timeouts in the same way.
        def blocking_request(*args, **kwargs):
            try:
                return self.client.blocking_request(*args, **kwargs)
            except RuntimeError as e:
                if not e.args[0].startswith('Request '):
                    raise

        # By default message identifiers should be enabled, and should start
        # counting at 1
        blocking_request(Message.request('watchdog'), timeout=0)
        blocking_request(Message.request('watchdog'), timeout=0)
        blocking_request(Message.request('watchdog'), timeout=0)
        # Extract Message object .mid attributes from the mock calls to
        # send_message
        mids = [args[0].mid              # arg[0] should be the Message() object
                for args, kwargs in self.client.send_message.call_args_list]
        self.assertEqual(mids, [b'1', b'2', b'3'])
        self.client.send_message.reset_mock()

        # Explicitly ask for no mid to be used
        blocking_request(
            Message.request('watchdog'), use_mid=False, timeout=0)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, None)

        # Ask for a specific mid to be used
        self.client.send_message.reset_mock()
        blocking_request(Message.request('watchdog', mid=42), timeout=0)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, b'42')

        ## Check situation for a katcpv4 server
        self.client._server_supports_ids = False

        # Should fail if an mid is passed
        reply, informs =  blocking_request(
            Message.request('watchdog', mid=42), timeout=0)
        self.assertFalse(reply.reply_ok())

        # Should fail if an mid is requested
        reply, informs = blocking_request(
            Message.request('watchdog'), use_mid=True, timeout=0)
        self.assertFalse(reply.reply_ok())

        # Should use no mid by default
        self.client.send_message.reset_mock()
        blocking_request(Message.request('watchdog'), timeout=0)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, None)

    def test_timeout(self):
        """Test calling blocking_request with a timeout."""
        reply, informs = self.client.blocking_request(
            Message.request("slow-command", "0.5"), timeout=0.001)
        self.assertFalse(reply.reply_ok())
        self.assertRegexpMatches(
            reply.arguments[1],
            br"Request slow-command timed out after 0\..* seconds.")

class TestCallbackClient(unittest.TestCase, TestUtilMixin):

    def setUp(self):
        self.addCleanup(self.stop_server_client)
        self.server = DeviceTestServer('', 0)
        self.server.start(timeout=0.1)

        host, port = self.server.bind_address

        self.client = katcp.CallbackClient(host, port)
        self.client.start(timeout=1)
        self.assertTrue(self.client.wait_protocol(timeout=1))

    def stop_server_client(self):
        if hasattr(self, 'client') and self.client.running():
            self.client.stop()
            self.client.join()
        if hasattr(self, 'server') and self.server.running():
            self.server.stop()
            self.server.join()

    def test_callback_request(self):
        """Test callback request."""

        watchdog_replies = []
        watchdog_replied = threading.Event()

        def watchdog_reply(reply):
            self.assertEqual(reply.name, "watchdog")
            self.assertEqual(reply.arguments, [b"ok"])
            watchdog_replies.append(reply)
            watchdog_replied.set()

        self.assertTrue(self.client.wait_protocol(0.5))
        self.client.callback_request(
            Message.request("watchdog"),
            reply_cb=watchdog_reply,
        )

        watchdog_replied.wait(0.5)
        self.assertTrue(watchdog_replies)

        help_replies = []
        help_informs = []
        help_replied = threading.Event()

        def help_reply(reply):
            self.assertEqual(reply.name, "help")
            self.assertEqual(reply.arguments, [b"ok", b"%d" % NUM_HELP_MESSAGES])
            self.assertEqual(len(help_informs), int(reply.arguments[1]))
            help_replies.append(reply)
            help_replied.set()

        def help_inform(inform):
            self.assertEqual(inform.name, "help")
            self.assertEqual(len(inform.arguments), 2)
            help_informs.append(inform)


        self.client.callback_request(
            Message.request("help"),
            reply_cb=help_reply,
            inform_cb=help_inform,
        )

        help_replied.wait(1)
        self.assertTrue(help_replied.isSet())
        help_replied.clear()
        help_replied.wait(0.05)   # Check if (unwanted) late help replies arrive
        self.assertFalse(help_replied.isSet())
        self.assertEqual(len(help_replies), 1)
        self.assertEqual(len(help_informs), NUM_HELP_MESSAGES)

    def test_callback_request_mid(self):
        ## Test that the client does the right thing with message identifiers

        # Wait for the client to detect the server protocol. Server should
        # support message identifiers
        self.assertTrue(self.client.wait_protocol(0.2))
        # Replace send_message so that we can check the message
        self.client.send_message = mock.Mock(wraps=self.client.send_message)

        # By default message identifiers should be enabled, and should start
        # counting at 1
        cb = counting_callback(number_of_calls=3)(lambda *x: x)
        self.client.callback_request(Message.request('watchdog'), reply_cb=cb)
        self.client.callback_request(Message.request('watchdog'), reply_cb=cb)
        self.client.callback_request(Message.request('watchdog'), reply_cb=cb)
        cb.assert_wait()
        # Extract Message object .mid attributes from the mock calls to
        # send_message
        mids = [args[0].mid              # args[0] should be the Message() object
                for args, kwargs in self.client.send_message.call_args_list]
        self.assertEqual(mids, [b'1', b'2', b'3'])
        self.client.send_message.reset_mock()

        # Explicitly ask for no mid to be used
        cb = counting_callback(number_of_calls=1)(lambda *x: x)
        self.client.callback_request(
            Message.request('watchdog'), use_mid=False, reply_cb=cb)
        cb.assert_wait()
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, None)

        # Ask for a specific mid to be used
        self.client.send_message.reset_mock()
        cb = counting_callback(number_of_calls=1)(lambda *x: x)
        self.client.callback_request(
            Message.request('watchdog', mid=42), reply_cb=cb)
        cb.assert_wait()
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, b'42')

        ## Check situation for a katcpv4 server
        self.client._server_supports_ids = False

        # Should fail if an mid is passed
        reply = [None]
        @counting_callback()
        def cb(msg):
            reply[0] = msg
        self.client.callback_request(
            Message.request('watchdog', mid=42), reply_cb=cb)
        cb.assert_wait()
        self.assertFalse(reply[0].reply_ok())

        # Should fail if an mid is requested
        reply = [None]
        cb.reset()
        self.client.callback_request(
            Message.request('watchdog'), use_mid=True, reply_cb=cb)
        cb.assert_wait()
        self.assertFalse(reply[0].reply_ok())

        # Should use no mid by default
        self.client.send_message.reset_mock()
        cb = counting_callback(number_of_calls=1)(lambda *x: x)
        self.client.callback_request(Message.request('watchdog'), reply_cb=cb)
        cb.assert_wait(timeout=1)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, None)

    def test_no_callback(self):
        """Test request without callback."""

        help_messages = []
        help_completed = threading.Event()

        def handle_help_message(client, msg):
            help_messages.append(msg)
            if msg.mtype == msg.REPLY:
                help_completed.set()

        self.client._inform_handlers["help"] = handle_help_message
        self.client._reply_handlers["help"] = handle_help_message
        # Set client._last_msg_id so we know that the ID is. Should be
        # _last_msg_id + 1
        self.client._last_msg_id = 0
        self.assertTrue(self.client.wait_protocol(0.2))
        self.client.callback_request(Message.request("help"))
        help_completed.wait(1)
        self.assertTrue(help_completed.isSet())

        self._assert_msgs_like(help_messages,
            [("#help[1] ", "")] * NUM_HELP_MESSAGES +
            [("!help[1] ok %d" % NUM_HELP_MESSAGES, "")])

    def test_timeout(self):
        self._test_timeout()

    def test_timeout_nomid(self, use_mid=False):
        self._test_timeout(use_mid=False)

    def _test_timeout(self, use_mid=None):
        """Test requests that timeout."""

        replies = []
        replied = threading.Event()
        informs = []
        timeout = 0.001

        @counting_callback()
        def reply_cb(msg):
            replies.append(msg)
            replied.set()

        def inform_cb(msg):
            informs.append(msg)

        self.assertTrue(self.client.wait_protocol(0.2))
        self.client.callback_request(
            Message.request("slow-command", "0.1"),
            use_mid=use_mid,
            reply_cb=reply_cb,
            inform_cb=inform_cb,
            timeout=timeout,
        )

        reply_cb.assert_wait(1)
        self.client.request(Message.request('cancel-slow-command'))
        msg = replies[0]
        self.assertEqual(msg.name, "slow-command")
        self.assertFalse(msg.reply_ok())
        self.assertRegexpMatches(msg.arguments[1],
                                 br"Request slow-command timed out after 0\..* seconds.")
        self.assertEqual(len(remove_version_connect(informs)), 0)
        self.assertEqual(len(replies), 1)

        del replies[:]
        del informs[:]
        reply_cb.reset()

        # test next request succeeds
        self.client.callback_request(
            Message.request("slow-command", "0.05"),
            reply_cb=reply_cb,
            inform_cb=inform_cb,
        )

        reply_cb.assert_wait()
        self.assertEqual(len(replies), 1)
        self.assertEqual(len(informs), 0)
        self.assertEqual([msg.name for msg in replies + informs],
                         ["slow-command"] * len(replies + informs))
        self.assertEqual([msg.arguments for msg in replies], [[b"ok"]])

    def test_timeout_nocb(self):
        """Test requests that timeout with no callbacks."""
        # Included to test https://katfs.kat.ac.za/mantis/view.php?id=1722
        # Situation can occur during a race between the timeout handler and the
        # receipt of a reply -- the reply can arrive after the timeout timer has
        # expired but before the request has been popped off the stack with
        # client._pop_async_request(). The normal request handler then pops off
        # the request first, resulting in the timeout handler getting a bunch of
        # None's. It should handle this gracefully.

        # Running the handler with a fake msg_id should have the same result as
        # running it after a real request has already been popped. The expected
        # result is that no assertions are raised.

        # NM 2014-09-26: This is probably no longer an issue with the tornado-based client
        # implementatin, but leaving the test for good measure
        f = Future()
        @gen.coroutine
        def cb():
            self.client._handle_timeout('fake_msg_id', time.time())
        self.client.ioloop.add_callback(lambda : gen.chain_future(cb(), f))
        f.result(timeout=1)


    def test_user_data(self):
        """Test callbacks with user data."""
        help_replies = []
        help_informs = []
        done = threading.Event()

        def help_reply(reply, x, y):
            self.assertEqual(reply.name, "help")
            self.assertEqual(x, 5)
            self.assertEqual(y, "foo")
            help_replies.append(reply)
            done.set()

        def help_inform(inform, x, y):
            self.assertEqual(inform.name, "help")
            self.assertEqual(x, 5)
            self.assertEqual(y, "foo")
            help_informs.append(inform)

        self.client.callback_request(
            Message.request("help"),
            reply_cb=help_reply,
            inform_cb=help_inform,
            user_data=(5, "foo"))

        done.wait(1)
        # Wait a bit longer to see if spurious replies arrive
        time.sleep(0.01)
        self.assertEqual(len(help_replies), 1)
        self.assertEqual(len(remove_version_connect(help_informs)),
                         NUM_HELP_MESSAGES)

    def test_fifty_thread_mayhem(self):
        """Test using callbacks from fifty threads simultaneously."""
        num_threads = 50
        # map from thread_id -> (replies, informs)
        results = {}
        # list of thread objects
        threads = []

        def reply_cb(reply, thread_id):
            results[thread_id][0].append(reply)
            results[thread_id][2].set()

        def inform_cb(inform, thread_id):
            results[thread_id][1].append(inform)

        def worker(thread_id, request):
            self.client.callback_request(
                request.copy(),
                reply_cb=reply_cb,
                inform_cb=inform_cb,
                user_data=(thread_id,),
            )

        request = Message.request("help")

        for thread_id in range(num_threads):
            results[thread_id] = ([], [], threading.Event())

        for thread_id in range(num_threads):
            thread = threading.Thread(target=worker, args=(thread_id, request))
            threads.append(thread)

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        for thread_id in range(num_threads):
            replies, informs, done = results[thread_id]
            done.wait(5.0)
            self.assertTrue(done.isSet())
            self.assertEqual(len(replies), 1)
            self.assertEqual(replies[0].arguments[0], b"ok")
            informs = remove_version_connect(informs)
            if len(informs) != NUM_HELP_MESSAGES:
                print(thread_id, len(informs))
                print([x.arguments[0] for x in informs])
            self.assertEqual(len(informs), NUM_HELP_MESSAGES)

    def test_blocking_request(self):
        """Test the callback client's blocking request."""
        reply, informs = self.client.blocking_request(
            Message.request("help"),
        )

        self.assertEqual(reply.name, "help")
        self.assertEqual(reply.arguments, [b"ok", b"%d" % NUM_HELP_MESSAGES])
        self.assertEqual(len(remove_version_connect(informs)), NUM_HELP_MESSAGES)

        reply, informs = self.client.blocking_request(
            Message.request("slow-command", "0.5"),
            timeout=0.001)

        self.assertEqual(reply.name, "slow-command")
        self.assertEqual(reply.arguments[0], b"fail")
        self.assertRegexpMatches(
            reply.arguments[1],
            br"Request slow-command timed out after 0\..* seconds.")

    def test_blocking_request_mid(self):
        ## Test that the blocking client does the right thing with message
        ## identifiers

        # Wait for the client to detect the server protocol. Server should
        # support message identifiers
        self.assertTrue(self.client.wait_protocol(1))
        # Replace send_message so that we can check the message
        self.client.send_message = mock.Mock()

        # By default message identifiers should be enabled, and should start
        # counting at 1
        self.client.blocking_request(Message.request('watchdog'), timeout=0)
        self.client.blocking_request(Message.request('watchdog'), timeout=0)
        self.client.blocking_request(Message.request('watchdog'), timeout=0)
        # Extract Message object .mid attributes from the mock calls to
        # send_message
        mids = [args[0].mid              # arg[0] should be the Message() object
                for args, kwargs in self.client.send_message.call_args_list]
        self.assertEqual(mids, [b'1', b'2', b'3'])
        self.client.send_message.reset_mock()

        # Explicitly ask for no mid to be used
        self.client.blocking_request(Message.request('watchdog'),
                                     use_mid=False, timeout=0)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, None)

        # Ask for a specific mid to be used
        self.client.send_message.reset_mock()
        self.client.blocking_request(Message.request(
            'watchdog', mid=42), timeout=0)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, b'42')

        ## Check situation for a katcpv4 server
        self.client._server_supports_ids = False

        # Should fail if an mid is passed
        reply, inform = self.client.blocking_request(Message.request(
                'watchdog', mid=42), timeout=0)
        self.assertFalse(reply.reply_ok())

        # Should fail if an mid is requested
        reply, inform = self.client.blocking_request(
            Message.request('watchdog'), use_mid=True, timeout=0)
        self.assertFalse(reply.reply_ok())

        # Should use no mid by default
        self.client.send_message.reset_mock()
        self.client.blocking_request(Message.request(
            'watchdog'), timeout=0)
        mid = self.client.send_message.call_args[0][0].mid
        self.assertEqual(mid, None)

    def test_request_fail_on_raise(self):
        """Test that the callback is called even if send_message raises
           KatcpClientError."""
        def raise_error(msg, timeout=None):
            raise katcp.KatcpClientError("Error %s" % msg.name)
        self.client.send_message = raise_error

        replies = []

        @counting_callback()
        def reply_cb(msg):
            replies.append(msg)

        self.client.callback_request(Message.request("foo"),
            reply_cb=reply_cb,
        )

        reply_cb.assert_wait()
        self.assertEqual(len(replies), 1)
        self.assertEqual(replies[0].name, "foo")
        self.assertEqual(replies[0].arguments, [b"fail", b"Error foo"])

    def test_stop_cleanup(self):
        self.client.wait_protocol(timeout=1)
        mid = 56
        future_reply = Future()
        self.client.ioloop.add_callback(
            lambda : gen.chain_future(self.client.future_request(Message.request(
                'slow-command', 1, mid=mid)), future_reply))
        # Force a disconnect
        self.client.stop()
        reply, informs = future_reply.result(timeout=1)
        self.assertEqual(reply, Message.reply(
            'slow-command', 'fail', 'Client stopped before reply was received', mid=mid))

class test_AsyncClientIntegrated(tornado.testing.AsyncTestCase, TestUtilMixin):
    def setUp(self):
        super(test_AsyncClientIntegrated, self).setUp()
        self.server = DeviceTestServer('', 0)
        self.server.set_ioloop(self.io_loop)
        self.server.set_concurrency_options(thread_safe=False, handler_thread=False)
        self.server.start()

        host, port = self.server.bind_address
        self.client = katcp.CallbackClient(host, port)
        self.client.set_ioloop(self.io_loop)
        self.client.start()

    @tornado.testing.gen_test
    def test_timeout_of_until_connected(self):
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield self.client.until_connected(timeout=0.0001)
        # Test for NOT timing out
        host, port = self.server.bind_address
        client2 = katcp.CallbackClient(host, port)
        client2.set_ioloop(self.io_loop)
        client2.start()
        yield client2.until_connected(timeout=0.5)
        self.assertTrue(client2.is_connected)

    @tornado.testing.gen_test
    def test_timeout_of_until_protocol(self):
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield self.client.until_protocol(timeout=0.0001)
        # Test for NOT timing out
        host, port = self.server.bind_address
        client2 = katcp.CallbackClient(host, port)
        client2.set_ioloop(self.io_loop)
        client2.start()
        yield client2.until_protocol(timeout=0.5)

    @tornado.testing.gen_test
    def test_future_request_simple(self):
        yield self.client.until_connected()
        reply, informs = yield self.client.future_request(Message.request('watchdog'))
        self.assertEqual(len(informs), 0)
        self.assertEqual(reply.name, "watchdog")
        self.assertEqual(reply.arguments, [b"ok"])

    @tornado.testing.gen_test
    def test_future_request_with_informs(self):
        yield self.client.until_connected()
        reply, informs = yield self.client.future_request(Message.request('help'))
        self.assertEqual(reply.name, "help")
        self.assertEqual(reply.arguments, [b"ok", b"%d" % NUM_HELP_MESSAGES])
        self.assertEqual(len(informs), NUM_HELP_MESSAGES)

    @tornado.testing.gen_test
    def test_disconnect_cleanup(self):
        yield self.client.until_protocol()
        mid = 55
        future_reply = self.client.future_request(Message.request(
            'slow-command', 1, mid=mid))
        # Force a disconnect
        self.client._disconnect()
        reply, informs = yield future_reply
        self.assertEqual(reply, Message.reply(
            'slow-command', 'fail', 'Connection closed before reply was received',
            mid=mid))

    @tornado.testing.gen_test
    def test_stop_cleanup(self):
        yield self.client.until_protocol()
        mid = 564
        future_reply = self.client.future_request(Message.request(
            'slow-command', 1, mid=mid))
        # Stop client
        self.client.stop()
        reply, informs = yield future_reply
        self.assertEqual(reply, Message.reply(
            'slow-command', 'fail', 'Client stopped before reply was received', mid=mid))

class test_AsyncClientIntegratedBase(TimewarpAsyncTestCase):
    def setUp(self):
        super(test_AsyncClientIntegratedBase, self).setUp()
        self.server = DeviceTestServer('localhost', 0)
        self.server.set_ioloop(self.io_loop)
        self.server.set_concurrency_options(thread_safe=False, handler_thread=False)
        self.server.start()

        host, port = self.server.bind_address
        logger.info('host, port: {}:{}'.format(host, port))
        self.client = katcp.CallbackClient(host, port)
        self.client.set_ioloop(self.io_loop)


class test_AsyncClientIntegrated(test_AsyncClientIntegratedBase):

    @tornado.testing.gen_test()
    def test_last_connect_time(self):
        self.server.stop()
        yield self.wake_ioloop()
        self.client.start()
        yield self.client.until_running()
        # Client is running, but server has stopped, so there should be no connection
        self.assertEqual(self.client.last_connect_time, None,
                         "last_connect_time should be 'None' before first connection")

        # Start up the server
        self.server.start()
        # and give the client time to connect
        yield self.client._waiting_to_retry.until_set()
        t0 = self.io_loop.time()
        self.set_ioloop_time(t0 + self.client.auto_reconnect_delay*1.1)
        # The connection should be made at the current ioloop time
        tc0 = self.io_loop.time()
        # Wait for the client to go through the motions of connecting
        yield self.client.until_connected()
        # Check that the correct connection time was stored
        self.assertEqual(self.client.last_connect_time, tc0)

        # Move time along and check that the last_connect_time is unchanged
        t1 = tc0 + 1.5
        self.set_ioloop_time(t1)
        yield self.wake_ioloop()
        # Connect time should not have changed since we did not reconnect
        self.assertEqual(self.client.last_connect_time, tc0)

        # Stop the server and check that last_connect_time is not affected
        self.server.stop()
        yield self.client._waiting_to_retry.until_set()
        self.assertFalse(self.client.is_connected())
        # Connect time should not have changed since we did not reconnect
        self.assertEqual(self.client.last_connect_time, tc0)

        # Restart the server and check that last_connect_time is updated
        self.server.start()
        t2 = t1 + 1.2
        self.set_ioloop_time(t2)
        yield self.client.until_connected()
        # Connect time should now be the current time
        self.assertEqual(self.client.last_connect_time, t2)


class test_AsyncClientTimeoutsIntegrated(test_AsyncClientIntegratedBase):
    def setUp(self):
        super(test_AsyncClientTimeoutsIntegrated, self).setUp()
        self.client.start()


    @tornado.testing.gen_test(timeout=10)
    # We are using time-warping, so the timeout should be longer than the fake-duration
    def test_future_request_default_timeout(self):
        # Test the default timeout of 5s
        yield self._test_timeout(5)

    @tornado.testing.gen_test()
    def test_future_request_change_default_timeout(self):
        self.client._request_timeout = 3
        yield self._test_timeout(3)

    @tornado.testing.gen_test()
    def test_future_request_request_timeout(self):
        yield self._test_timeout(1, set_request_timeout=True)

    @gen.coroutine
    def _test_timeout(self, timeout, set_request_timeout=False):
        request_timeout = timeout if set_request_timeout else None
        yield self.client.until_connected()
        t0 = self.io_loop.time()
        reply_future = self.client.future_request(Message.request('slow-command', timeout + 1),
                                                  timeout=request_timeout)
        # Warp to just before the timeout expires and check that the future is not yet
        # resolved
        self.set_ioloop_time(t0 + timeout*0.9999)
        yield self.wake_ioloop()
        self.assertFalse(reply_future.done())
        # Warp to just after the timeout expires, and check that it gives us a timeout
        # error reply
        self.set_ioloop_time(t0 + timeout*1.0001)
        yield self.wake_ioloop()
        self.assertTrue(reply_future.done())
        reply, informs = reply_future.result()
        self.assertFalse(reply.reply_ok())
        self.assertRegexpMatches(
            reply.arguments[1],
            br"Request slow-command timed out after .* seconds.")



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details
from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import logging
import sys
import textwrap
import threading
import time

from builtins import object
from concurrent.futures import Future, TimeoutError
from functools import wraps

import tornado.ioloop

from _thread import get_ident as get_thread_ident
from tornado import gen
from tornado.concurrent import Future as tornado_Future

if sys.version_info[0] == 3:
    from tornado.platform.asyncio import AsyncIOLoop

from katcp.object_proxies import ObjectWrapper


log = logging.getLogger(__name__)


def with_relative_timeout(timeout, future, io_loop=None):
    return gen.with_timeout(timeout + time.time(), future, io_loop)


class IOLoopManager(object):
    """Manages an IOLoop instance, optionally in a separate thread."""

    def __init__(self, managed_default=True, logger=log):
        # True if we manage the ioloop. Will be updated by self.set_ioloop()
        self._ioloop_managed = managed_default
        self._logger = logger
        # Thread object that a managed ioloop is running in
        self._ioloop_thread = None
        self._ioloop = None
        # Event that indicates that the ioloop is running.
        self._running = threading.Event()
        self._start_lock = threading.Lock()
        self._daemonize = False

    @property
    def managed(self):
        return self._ioloop_managed

    def get_ioloop(self):
        if not self._ioloop:
            if self._ioloop_managed:
                if sys.version_info[0] == 3:
                    self.set_ioloop(AsyncIOLoop())
                else:
                    self.set_ioloop(tornado.ioloop.IOLoop())
            else:
                self.set_ioloop(tornado.ioloop.IOLoop.current())
        return self._ioloop

    def set_ioloop(self, ioloop, managed=None):
        if managed is not None:
            self._ioloop_managed = managed
        if self._ioloop:
            raise RuntimeError('IOLoop instance already set')
        self._ioloop = ioloop

    def _run_managed_ioloop(self):
        assert self._ioloop_managed

        def run_ioloop():
            try:
                self._ioloop.start()
                self._ioloop.close(all_fds=True)
            except Exception:
                self._logger.error('Error running tornado IOloop: ',
                                   exc_info=True)
            finally:
                ioloop = self._ioloop
                self._ioloop = None
                self._running.clear()
                self._logger.info('Managed tornado IOloop {0} stopped'
                                  .format(ioloop))

        with self._start_lock:
            t = threading.Thread(target=run_ioloop)
            t.setDaemon(self._daemonize)
            try:
                if self._ioloop_thread.isAlive():
                    raise RuntimeError('Seems that managed ioloop has already been '
                                       'started, can only restart after stop()')
            except AttributeError:
                pass
            self._ioloop_thread = t
            self._ioloop_thread.start()

    def setDaemon(self, daemonic):
        """Set daemonic state of the managed ioloop thread to True / False

        Calling this method for a non-managed ioloop has no effect. Must be called before
        start(), or it will also have no effect
        """
        self._daemonize = bool(daemonic)

    def start(self, timeout=None):
        """Start managed ioloop thread, or do nothing if not managed.

        If a timeout is passed, it will block until the the event loop is alive
        (or the timeout expires) even if the ioloop is not managed.

        """
        if not self._ioloop:
            raise RuntimeError('Call get_ioloop() or set_ioloop() first')

        self._ioloop.add_callback(self._running.set)

        if self._ioloop_managed:
            self._run_managed_ioloop()
        else:            #  TODO this seems inconsistent with what the docstring describes
            self._running.set()

        if timeout:
            return self._running.wait(timeout)

    def stop(self, timeout=None, callback=None):
        """Stop ioloop (if managed) and call callback in ioloop before close.

        Parameters
        ----------
        timeout : float or None
            Seconds to wait for ioloop to have *started*.

        Returns
        -------
        stopped : thread-safe Future
            Resolves when the callback() is done

        """
        if timeout:
            self._running.wait(timeout)

        stopped_future = Future()

        @gen.coroutine
        def _stop():
            if callback:
                try:
                    yield gen.maybe_future(callback())
                except Exception:
                    self._logger.exception('Unhandled exception calling stop callback')
            if self._ioloop_managed:
                self._logger.info('Stopping ioloop {0!r}'.format(self._ioloop))
                # Allow ioloop to run once before stopping so that callbacks
                # scheduled by callback() above get a chance to run.
                yield gen.moment
                self._ioloop.stop()
            self._running.clear()

        try:
            self._ioloop.add_callback(
                lambda: gen.chain_future(_stop(), stopped_future))
        except AttributeError:
            # Probably we have been shut-down already
            pass

        return stopped_future

    def join(self, timeout=None):
        """Join managed ioloop thread, or do nothing if not managed."""
        if not self._ioloop_managed:
            # Do nothing if the loop is not managed
            return
        try:
            self._ioloop_thread.join(timeout)
        except AttributeError:
            raise RuntimeError('Cannot join if not started')

class IOLoopThreadWrapper(object):
    default_timeout = None

    def __init__(self, ioloop=None):
        self.ioloop = ioloop = ioloop or tornado.ioloop.IOLoop.current()
        self._thread_id = None
        ioloop.add_callback(self._install)

    def call_in_ioloop(self, fn, args, kwargs, timeout=None):
        timeout = timeout or self.default_timeout
        if get_thread_ident() == self._thread_id:
            raise RuntimeError("Cannot call a thread-wrapped object from the ioloop")
        future, tornado_future = Future(), tornado_Future()
        self.ioloop.add_callback(
            self._ioloop_call, future, tornado_future, fn, args, kwargs)
        try:
            # Use the thread-safe future to block
            return future.result(timeout)
        except TimeoutError:
            raise
        except Exception:
            # If we have an exception use the tornado future instead since it
            # will print a nicer traceback.
            tornado_future.result()
            # Should never get here since the tornado future should raise
            assert False, 'Tornado Future should have raised'

    def decorate_callable(self, callable_):
        """Decorate a callable to use call_in_ioloop"""
        @wraps(callable_)
        def decorated(*args, **kwargs):
            # Extract timeout from request itself or use default for ioloop wrapper
            timeout = kwargs.get('timeout')
            return self.call_in_ioloop(callable_, args, kwargs, timeout)

        decorated.__doc__ = '\n\n'.join((
"""Wrapped async call. Will call in ioloop.

This call will block until the original callable has finished running on the ioloop, and
will pass on the return value. If the original callable returns a future, this call will
wait for the future to resolve and return the value or raise the exception that the future
resolves with.

Original Callable Docstring
---------------------------
""",
            textwrap.dedent(decorated.__doc__ or '')))

        return decorated

    def _install(self):
        self._thread_id = get_thread_ident()

    def _ioloop_call(self, future, tornado_future, fn, args, kwargs):
        gen.chain_future(tornado_future, future)
        try:
            result_future = gen.maybe_future(fn(*args, **kwargs))
            gen.chain_future(result_future, tornado_future)
        except Exception:
            tornado_future.set_exc_info(sys.exc_info())

class ThreadSafeMethodAttrWrapper(ObjectWrapper):
    # Attributes must be in the class definition, or else they will be
    # proxied to __subject__
    _ioloop_wrapper = None

    def __init__(self, subject, ioloop_wrapper):
        self._ioloop_wrapper = ioloop_wrapper
        super(ThreadSafeMethodAttrWrapper, self).__init__(subject)

    def __getattr__(self, attr):
        val = super(ThreadSafeMethodAttrWrapper, self).__getattr__(attr)
        if callable(val):
            return self._ioloop_wrapper.decorate_callable(val)
        else:
            return val

    def _getattr(self, attr):
        return self._ioloop_wrapper.call_in_ioloop(getattr, (self.__subject__,
            attr), {})
